#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2016 Tom Samstag
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Unittest that tests person-specific filter rules
"""
import unittest
import os
from time import perf_counter
import inspect

from ....filters import reload_custom_filters
reload_custom_filters()
from ....db.utils import import_as_dict
from ....filters import GenericFilter, CustomFilters
from ....const import DATA_DIR
from ....user import User

from ..person import (
    Disconnected, Everyone, FamilyWithIncompleteEvent, HasAddress,
    HasAlternateName, HasAssociation, HasBirth, HasDeath, HasEvent,
    HasCommonAncestorWith, HasCommonAncestorWithFilterMatch,
    HasFamilyAttribute, HasFamilyEvent, HasIdOf, HasLDS,
    HasNameOf, HasNameOriginType, HasNameType, HasNickname, HasRelationship,
    HasSoundexName, HasSourceOf, HasTextMatchingRegexpOf, HasUnknownGender,
    HaveAltFamilies, HaveChildren, HavePhotos, IncompleteNames,
    IsAncestorOfFilterMatch, IsBookmarked, IsChildOfFilterMatch,
    IsDescendantFamilyOf, IsDescendantFamilyOfFilterMatch,
    IsDescendantOfFilterMatch, IsDefaultPerson, IsDescendantOf,
    IsDuplicatedAncestorOf, IsFemale,
    IsLessThanNthGenerationAncestorOf,
    IsLessThanNthGenerationAncestorOfDefaultPerson,
    IsLessThanNthGenerationAncestorOfBookmarked, IsMale,
    IsMoreThanNthGenerationAncestorOf, IsMoreThanNthGenerationDescendantOf,
    IsParentOfFilterMatch, IsRelatedWith, IsSiblingOfFilterMatch,
    IsSpouseOfFilterMatch, IsWitness, MissingParent, MultipleMarriages,
    NeverMarried, NoBirthdate, NoDeathdate, PeoplePrivate, PeoplePublic,
    PersonWithIncompleteEvent, ProbablyAlive, RegExpName,
    RelationshipPathBetweenBookmarks,
)

TEST_DIR = os.path.abspath(os.path.join(DATA_DIR, "tests"))
EXAMPLE = os.path.join(TEST_DIR, "example.gramps")

class BaseTest(unittest.TestCase):
    """
    Person rule tests.
    """

    @classmethod
    def setUpClass(cls):
        """
        Import example database.
        """
        cls.db = import_as_dict(EXAMPLE, User())

    def filter_with_rule(self, rule, l_op='and', invert=False,
                         baserule=None, base_l_op='and', base_invert=False,
                         base_name='Base'):
        """
        Apply a filter with the given rule.  'baserule' can be used to stack
        filters when using filters that are of 'offiltermatch' type.
        """
        if baserule:
            filter_ = GenericFilter()
            if isinstance(baserule, list):
                filter_.set_rules(baserule)
            else:
                filter_.add_rule(baserule)
            filter_.set_logical_op(base_l_op)
            filter_.set_invert(base_invert)
            filter_.set_name(base_name)
            filters = CustomFilters.get_filters_dict('Person')
            filters[base_name] = filter_
        filter_ = GenericFilter()
        if isinstance(rule, list):
            filter_.set_rules(rule)
        else:
            filter_.add_rule(rule)
        filter_.set_logical_op(l_op)
        filter_.set_invert(invert)
        stime = perf_counter()
        results = filter_.apply(self.db)
        if __debug__:
            rulename = inspect.stack()[1][3]
            print("%s: %.2f\n" % (rulename, perf_counter() - stime))
        return set(results)

    def test_Complex_1(self):
        """ Test with two ancestor trees in base filter, and a complex
        'or' of a descendent tree, sibling of, """
        rule1 = IsLessThanNthGenerationAncestorOf(['I0005', 10])
        rule2 = IsLessThanNthGenerationAncestorOf(['I0006', 10])
        ruleA = IsDescendantOfFilterMatch(['Base'])
        ruleB = IsSiblingOfFilterMatch(['Base'])
        ruleC = IsSpouseOfFilterMatch(['Base'])
        res = self.filter_with_rule([ruleA, ruleB, ruleC], l_op='or',
                                    baserule=[rule1, rule2], base_l_op='or')
        self.assertEqual(len(res), 1277)

    def test_IsDescendantOfFilterMatch(self):
        """ Test the rule with two persons in base filter """
        rule1 = HasIdOf(['I0610'])
        rule2 = HasIdOf(['I1804'])
        rule = IsDescendantOfFilterMatch(['Base'])
        res = self.filter_with_rule(rule, baserule=[rule1, rule2],
                                    base_l_op='or')
        self.assertEqual(len(res), 120)

    def test_IsDescendantFamilyOfFilterMatch(self):
        """ Test the rule with two persons in base filter """
        rule1 = HasIdOf(['I0610'])
        rule2 = HasIdOf(['I1804'])
        rule = IsDescendantFamilyOfFilterMatch(['Base'])
        res = self.filter_with_rule(rule, baserule=[rule1, rule2],
                                    base_l_op='or')
        self.assertEqual(len(res), 171)

    def test_IsDescendantFamilyOf(self):
        """ Test the rule """
        rule = IsDescendantFamilyOf(['I0610', 1])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 118)

    def test_IsDescendantOf(self):
        """ Test the rule """
        rule = IsDescendantOf(['I0610', 0])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 85)

    def test_IsMoreThanNthGenerationDescendantOf(self):
        """ Test the rule """
        rule = IsMoreThanNthGenerationDescendantOf(['I0610', 3])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 83)

    def test_IsLessThanNthGenerationAncestorOfDefaultPerson(self):
        """ Test the rule """
        rule = IsLessThanNthGenerationAncestorOfDefaultPerson([5])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 7)

    def test_IsLessThanNthGenerationAncestorOfBookmarked(self):
        """ Test the rule """
        rule = IsLessThanNthGenerationAncestorOfBookmarked([5])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 8)

    def test_IsParentOfFilterMatch(self):
        """ Test the rule with two persons in base filter """
        rule1 = HasIdOf(['I0006'])
        rule2 = HasIdOf(['I0010'])
        rule = IsParentOfFilterMatch(['Base'])
        res = self.filter_with_rule(rule, baserule=[rule1, rule2],
                                    base_l_op='or')
        self.assertEqual(len(res), 4)

    def test_IsSiblingOfFilterMatch(self):
        """ Test the rule with two persons in base filter """
        rule1 = HasIdOf(['I0006'])
        rule2 = HasIdOf(['I0010'])
        rule = IsSiblingOfFilterMatch(['Base'])
        res = self.filter_with_rule(rule, baserule=[rule1, rule2],
                                    base_l_op='or')
        self.assertEqual(len(res), 7)

    def test_IsSpouseOfFilterMatch(self):
        """ Test the rule with two persons in base filter """
        rule1 = HasIdOf(['I0006'])
        rule2 = HasIdOf(['I0010'])
        rule = IsSpouseOfFilterMatch(['Base'])
        res = self.filter_with_rule(rule, baserule=[rule1, rule2],
                                    base_l_op='or')
        self.assertEqual(len(res), 2)

    def test_isancestoroffiltermatch(self):
        """ Test the rule with two persons in base filter """
        rule1 = HasIdOf(['I0006'])
        rule2 = HasIdOf(['I0005'])
        rule = IsAncestorOfFilterMatch(['Base'])
        res = self.filter_with_rule(rule, baserule=[rule1, rule2],
                                    base_l_op='or')
        self.assertEqual(len(res), 431)

    def test_HasCommonAncestorWithFilterMatch(self):
        """ Test the rule with two persons in base filter """
        rule1 = HasIdOf(['I0006'])
        rule2 = HasIdOf(['I0005'])
        rule = HasCommonAncestorWithFilterMatch(['Base'])
        res = self.filter_with_rule(rule, baserule=[rule1, rule2],
                                    base_l_op='or')
        self.assertEqual(len(res), 1415)

    def test_IsChildOfFilterMatch(self):
        """ Test the rule with two persons in base filter """
        rule1 = HasIdOf(['I0006'])
        rule2 = HasIdOf(['I0010'])
        rule = IsChildOfFilterMatch(['Base'])
        res = self.filter_with_rule(rule, baserule=[rule1, rule2],
                                    base_l_op='or')
        self.assertEqual(len(res), 11)

    def test_HasAddress(self):
        """
        Test HasAddress rule.
        """
        rule = HasAddress([0, 'greater than'])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 1)

    def test_HasAssociation(self):
        """
        Test rule.
        """
        rule = HasAssociation([0, 'greater than'])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 2)

    def test_HasBirth(self):
        """
        Test rule.
        """
        rule = HasBirth(['between 1600 and 1700', 'akron', ''])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 2)

    def test_HasDeath(self):
        """
        Test HasDeath rule.
        """
        rule = HasDeath(['between 1600 and 1700', 'ashtabula', ''])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 2)

    def test_HasEvent(self):
        """
        Test rule.
        """
        rule = HasEvent(['Birth', 'between 1600 and 1700', 'akron',
                         '', '', 1])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 2)

    def test_HasFamilyAttribute(self):
        """
        Test rule.
        """
        rule = HasFamilyAttribute([5, '8'])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 2)

    def test_HasFamilyEvent(self):
        """
        Test rule.
        """
        rule = HasFamilyEvent(['Marriage', 'after 1900', 'craw', ''])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 4)

    def test_HavePhotos(self):
        """
        Test rule.
        """
        rule = HavePhotos([0, 'greater than'])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 5)

    def test_HasLDS(self):
        """
        Test rule.
        """
        rule = HasLDS([0, 'greater than'])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 1)

    def test_HasNameOriginType(self):
        """
        Test rule.
        """
        rule = HasNameOriginType(['Patrilineal'])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 9)

    def test_HasNameType(self):
        """
        Test rule.
        """
        rule = HasNameType(['Married Name'])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 1)

    def test_HasRelationship(self):
        """
        Test rule.
        """
        rule = HasRelationship([0, 'Married', 0])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 1377)

    def test_HasTextMatchingRegexpOf(self):
        """
        Test rule.
        """
        rule = HasTextMatchingRegexpOf(['.*(Dahl|Akron|Smith|Attic|'
                                        'of Lessard).*', False],
                                       use_regex=True)
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 28)

    def test_IsMoreThanNthGenerationAncestorOf(self):
        """
        Test rule.
        """
        rule = IsMoreThanNthGenerationAncestorOf(['I0005', 3])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 322)

    def test_IsWitness(self):
        """
        Test rule.
        """
        rule = IsWitness(['Marriage'])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 1)

    def test_ProbablyAlive(self):
        """
        Test rule.
        """
        rule = ProbablyAlive(['1900'])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 766)

    def test_RegExpName(self):
        """
        Test rule.
        """
        rule = RegExpName(['.*(Garc|Amy).*'], use_regex=True)
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 3)


    def test_disconnected(self):
        """
        Test Disconnected rule.
        """
        rule = Disconnected([])
        self.assertEqual(self.filter_with_rule(rule), set([
            '0PBKQCXHLAEIB46ZIA', 'QEVJQC04YO01UAWJ2N', 'UT0KQCMN7PC9XURRXJ',
            'MZAKQCKAQLIQYWP5IW', 'Y7BKQC9CUXWQLGLPQM', 'OBBKQC8NJM5UYBO849',
            'NPBKQCKEF0G7T4H312', '423KQCGLT8UISDUM1Q', '8S0KQCNORIWDL0X8SB',
            'AP5KQC0LBXPM727OWB', 'AREKQC0VPBHNZ5R3IO', 'KU0KQCJ0RUTJTIUKSA',
            'VC4KQC7L7KKH9RLHXN', '0P3KQCRSIVL1A4VJ19', 'PK6KQCGEL4PTE720BL',
            'YIKKQCSD2Z85UHJ8LX', 'KY8KQCMIH2HUUGLA3R', 'RD7KQCQ24B1N3OEC5X',
            'NV0KQC7SIEH3SVDPP1', 'KIKKQCU2CJ543TLM5J', 'AT0KQC4P3MMUCHI3BK',
            'J6BKQC1PMNBAYSLM9U', 'IXXJQCLKOUAJ5RSQY4', 'U4ZJQC5VR0QBIE8DU',
            'F7BKQC4NXO9R7XOG2W', '7U0KQC6PGZBNQATNOT', '78AKQCI05U36T3E82O',
            'H1GKQCWOUJHFSHXABA', 'ZWGKQCRFZAPC5PYJZ1', 'EZ0KQCF3LSM9PRSG0K',
            'FHKKQC963NGSY18ZDZ', 'FJ9KQCRJ3RGHNBWW4S', 'S2EKQC9F4UR4R71IC3',
            '1XBKQCX019BKJ0M9IH', 'Z62KQC706L0B0WTN3Q', 'O7EKQCEVZ7FBEWMNWE',
            'XY8KQCULFPN4SR915Q', 'WQDKQCEULSD5G9XNFI', '2Z0KQCSWKVFG7RPFD8',
            '26BKQC0SJIJOH02H2A', '262KQCH2RQKN0CBRLF', 'P5ZJQCMKO7EYV4HFCL',
            'KXBKQC52JO3AP4GMLF', '9IFKQC60JTDBV57N6S', 'TQ0KQCZ8LA7X9DIEAN',
            'BAXJQCORQA5Q46FCDG', 'VR0KQC7LVANO83AL35', '75CKQC4T617U2E5T5Y',
            'LCTKQCZU3F94CEFSOM', 'WJYJQCPNJJI5JN07SD', '3N6KQC6BE5EIXTRMDL',
            'CM5KQCD57I15GKLAMB', 'cccbffffd3e69819cd8',
            'BJKKQCVDA66528PDAU', 'QS0KQCLMIZFI8ZDLM3', 'UW0KQCRHBIYMA8LPZD',
            'GJ7KQC7APJSAMHEK5Q', '711KQCDXOQWB3KDWEP', 'PY0KQC77AJ3457A6C2',
            'WZ0KQCYVMEJHDR4MV2', '28EKQCQGM6NLLWFRG7', 'E33KQCRREJALRA715H',
            '8HKKQCTEJAOBVH410L', 'IO6KQC70PMBQUDNB3L', '1YBKQCWRBNB433NEMH',
            'M01KQCF7KUWCDY67JD', 'CR0KQCOMV2QPPC90IF', '85ZJQCMG38N7Q2WKIK',
            'I9GKQCERACL8UZF2PY', 'BY0KQCOZUK47R2JZDE', '7W0KQCYDMD4LTSY5JL',
            'A0YJQC3HONEKD1JCPK', 'd5839c13b0541b7b8e6',
            ]))

    def test_everyone(self):
        """
        Test Everyone rule.
        """
        rule = Everyone([])
        self.assertEqual(len(self.filter_with_rule(rule)),
                         self.db.get_number_of_people())

    def test_familywithincompleteevent(self):
        """
        Test FamilyWithIncompleteEvent rule.
        """
        rule = FamilyWithIncompleteEvent([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 775)

    def test_hasalternatename(self):
        """
        Test HasAlternateName rule.
        """
        rule = HasAlternateName([])
        self.assertEqual(self.filter_with_rule(rule), set([
            '46WJQCIOLQ0KOX2XCC', 'GNUJQCL9MD64AM56OH',
            ]))

    def test_commonancestor_empty(self):
        """
        Test empty HasCommonAncestorWith rule.
        """
        rule = HasCommonAncestorWith([''])
        self.assertEqual(self.filter_with_rule(rule), set([
            ]))

    def test_commonancestor_nonmatching(self):
        """
        Test non-matching HasCommonAncestorWith rule.
        """
        rule = HasCommonAncestorWith(['I0000'])
        self.assertEqual(self.filter_with_rule(rule), set([
            'd5839c1237765987724'
            ]))

    def test_commonancestor_irregular(self):
        """
        Test irregular HasCommonAncestorWith rule.
        """
        rule = HasCommonAncestorWith(['ABCDEFG'])
        self.assertEqual(self.filter_with_rule(rule), set([
            ]))

    def test_commonancestor_matching(self):
        """
        Test matching HasCommonAncestorWith rule.
        """
        rule = HasCommonAncestorWith(['I0044'])
        self.assertEqual(self.filter_with_rule(rule), set([
            'GNUJQCL9MD64AM56OH', 'SOFKQCBYAO18OWC0CS', 'EMEKQC02EOUF8H0SHM',
            '3EXJQCVWOSQFGLYB6H', 'EMTJQCQU6TL4WAVSE4', 'QUEKQCZL61S8BJJ388',
            'MKEKQCSBQGAVHAPCQT', 'MUFKQCMXUJ07MCDUNI', 'DBXJQCJCEZMO17WZ89',
            'ORFKQC4KLWEGTGR19L', 'MG5KQC6ZKSVO4A63G2', 'N26KQCF3ASHMZ0HEW6',
            'GNWJQC9NLVF2MZLHU9', 'ZFXJQCHAD8SLZZ7KRP', '44WJQCLCQIPZUB0UH',
            'B8TJQC53HJXOGXK8F7', 'D3WJQCCGV58IP8PNHZ', '3LEKQCRF3FD2E1H73I',
            'F06KQCZY1I4H4IFZM', 'VMTJQC49IGKLG2EQ5', '9BXKQC1PVLPYFMD6IX',
            'H1DKQC4YGZ5A61FGS', '1GWJQCGOOZ8FJW3YK9', 'S16KQCX8XUO3EEL85N',
            'OREKQCF34YE89RL8S6', 'RU5KQCQTPC9SJ5Q1JN', 'GYFKQCPH8Q0JDN94GR',
            '9QFKQC54ET79K2SD57', 'MLEKQCH64557K610VR', 'AWFKQCJELLUWDY2PD3',
            'ZDWJQC7TMS2AWAVF2Y', 'VJFKQCFO7WESWPNKHE', 'LV5KQCJCCR0S3DN5WW',
            'CDTJQCVTVX7CNMY9YU', 'OX5KQCKE3I94MEPDC', 'JF5KQC2L6ABI0MVD3E',
            'CH5KQCIEXSN1J5UEHB', '4JEKQC22K5UTH9QHCU', 'EPFKQCETTDTEL3PYIR',
            'D16KQCIZS56HVPW6DA', '2TEKQCTSCRL4Z2AUHE', '3WEKQCHXRH61E3CIKB',
            'TDTJQCGYRS2RCCGQN3', 'SMWJQCXQ6I2GEXSPK9', 'PXFKQCXEHJX3W1Q1IV',
            'Q9TJQCXDL1599L2B2Z', 'BFXJQCF1JBOXPRW2OS', '6TFKQCUTO94WB2NHN',
            'FNEKQCO239QSNK0R78', '3RFKQCNKMX9HVLNSLW', 'W2DKQCV4H3EZUJ35DX',
            '5IEKQCN37EFBK9EBUD', 'LW5KQCXSXRC2XV3T3D', 'ZNEKQCULV911DIXBK3',
            '35WJQC1B7T7NPV8OLV', 'MPEKQC6TIP3SP1YF7I', 'DMFKQC5MHGYC6503F2',
            '3KEKQC45RL87D4ZG86', 'KLTJQC70XVZJSPQ43U', 'LVEKQCP09W7JNFDAFC',
            'DPUJQCUYKKDPT78JJV', 'JDXJQCR5L0NTR21SQA', 'UAXJQC6HC354V7Q6JA',
            'XBXJQCS4QY316ZGHRN', 'HCXJQCRKB4K65V1C07', '66TJQC6CC7ZWL9YZ64',
            'XNFKQC6DN59LACS9IU', 'LL5KQCG687Y165GL5P', '7X5KQC9ABK4T6AW7QF',
            'HKTJQCIJD8RK9RJFO1', '1LTJQCYQI1DXBLG6Z', '0FWJQCLYEP736P3YZK',
            '0DXJQC1T8P3CQKZIUO', 'ISEKQC97YI74A9VKWC', 'KGXJQCBQ39ON9VB37T',
            'BZ5KQCD4KFI3BTIMZU', '0HEKQCLINMQS4RB7B8', 'BBTJQCNT6N1H4X6TL4',
            'COFKQCUXC2H4G3QBYT', 'DI5KQC3CLKWQI3I0CC', 'T8TJQCWWI8RY57YNTQ',
            '46WJQCIOLQ0KOX2XCC', 'OEXJQCQJHF2BLSAAIS', 'GNFKQCH8AFJRJO9V4Y',
            '8LFKQCQWXTJQJR4CXV', 'IGWJQCSVT8NXTFXOFJ', '3PEKQC8ZDCYTSSIKZ9',
            '5UEKQC8N8NEPSWU1QQ', 'NK5KQC1MAOU2BP35ZV', 'UZFKQCIHVT44DC9KGH',
            'JJ5KQC83DT7VDMUYRQ', '626KQC7C08H3UTM38E', 'XIFKQCLQOY645QTGP7',
            'HEWJQCWQQ3K4BNRLIO', 'HDWJQCT361VOV2PQLP', 'XFKKQCGA4DVECEB48E',
            'KWEKQCTNIIV9BROFFG',
            ]))

    def test_hasnickname(self):
        """
        Test HasNickname rule.
        """
        rule = HasNickname([])
        self.assertEqual(self.filter_with_rule(rule), set([
            'cc8205d883763f02abd', 'GNUJQCL9MD64AM56OH',
            'Q8HKQC3VMRM1M6M7ES',
            ]))

    def test_hasunknowngender(self):
        """
        Test HasUnknownGender rule.
        """
        rule = HasUnknownGender([])
        self.assertEqual(self.filter_with_rule(rule), set([
            'OJOKQC83Y1EDBIMLJ6', '8BHKQCFK9UZFRJYC2Y', 'PGFKQC1TUQMXFAMLMB',
            'IHOKQCECRZYQDKW6KF', '8HKKQCTEJAOBVH410L', 'AGFKQCO358R18LNJYV',
            '1ENKQCBPFZTAQJSP4O', 'NUWKQCO7TVAOH0CHLV', 'P5IKQC88STY3FNTFZ3',
            '7GXKQCMVFU8WR1LKZL', 'LGXKQCJ5OP6MKF9QLN', 'XNFKQC6DN59LACS9IU',
            '7IOKQC1NVGUI1E55CQ', '57PKQCFAWY7AM3JS4M', 'BNXKQCEBXC1RCOGJNF',
            'TFFKQC1RMG8RRADKDH', 'FHKKQC963NGSY18ZDZ', 'WMXKQCDUJ4JKQQYCR7',
            'PBHKQCHOAGTECRKT9L', 'OFXKQC8W0N3N6JP6YQ',
            ]))

    def test_hassourceof_empty(self):
        """
        Test empty HasSourceOf rule.
        """
        # when run with an empty string finds people with no sourc citations
        rule = HasSourceOf([''])
        self.assertEqual(self.filter_with_rule(rule), set([
            'cc82060512042f67e2c', 'cc8205d87831c772e87',
            'cc82060516c6c141500', 'cc8205d87fd529000ff',
            'cc82060504445ab6deb', 'cc8205d887376aacba2',
            'cccbffffd3e69819cd8', 'cc8205d87c20350420b',
            'cc8206050e541f79f92', 'cc8205d883763f02abd',
            'cc8206050980ea622d0', 'cc8205d872f532ab14e',
            'd5839c132b11d9e3632', 'd583a5ba0d50afbbaaf',
            'd5839c1352c64b924d9', 'd583a5b9fc864e3bf4e',
            'd583a5ba1bd083ce4c2', 'd583a5b9df71bceb48c',
            'd583a5b9ced473a7e6a', 'd583a5ba2bc7b9d1388',
            'd5839c12fec09785f6a', 'd5839c1237765987724',
            'd5839c137b3640ad776', 'd5839c126d11a754f46',
            'd5839c12d3b4d5e619b', 'd5839c13380462b246f',
            'd5839c12e9e08301ce2', 'd5839c1366b21411fb4',
            'd5839c13a282b51dd0d', 'd5839c12ac91650a72b',
            'd583a5b9edf6cb5d8d5', 'd583a5ba4be3acdd312',
            'd5839c131d560e06bac', 'd5839c13b0541b7b8e6',
            'd5839c1388e3ab6c87c', 'd583a5ba5ca6b698463',
            'd583a5ba3bc48c2002c', 'd583a5b90777391ea9a',
            ]))

    def test_hassourceof_nonmatching(self):
        """
        Test non-matching HasSourceOf rule.
        """
        rule = HasSourceOf(['S0004'])
        self.assertEqual(self.filter_with_rule(rule), set([
            ]))

    def test_hassourceof_irregular(self):
        """
        Test irregular HasSourceOf rule.
        """
        rule = HasSourceOf(['ABCDEFG'])
        self.assertEqual(self.filter_with_rule(rule), set([
            ]))

    def test_hassourceof_matching(self):
        """
        Test matching HasSourceOf rule.
        """
        rule = HasSourceOf(['S0000'])
        self.assertEqual(self.filter_with_rule(rule), set([
            'GNUJQCL9MD64AM56OH',
            ]))


    def test_havealtfamilies(self):
        """
        Test HaveAltFamilies rule.
        """
        rule = HaveAltFamilies([])
        self.assertEqual(self.filter_with_rule(rule), set([
            'CH5KQCIEXSN1J5UEHB', 'MG5KQC6ZKSVO4A63G2',
            ]))

    def test_havechildren(self):
        """
        Test HaveChildren rule.
        """
        rule = HaveChildren([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 905)

    def test_incompletenames(self):
        """
        Test IncompleteNames rule.
        """
        rule = IncompleteNames([])
        self.assertEqual(self.filter_with_rule(rule), set([
            'IHOKQCECRZYQDKW6KF', 'cc82060504445ab6deb',
            'LCXKQCQZH5EH56NTCD', 'cc8205d87831c772e87',
            '3RFKQCNKMX9HVLNSLW', 'cc8205d87fd529000ff',
            'B1UKQCBR49WB3134PN', '0TTKQCXXY59OCDPLV3',
            'F3UKQC7ZV3EYVWTZ8O', '1MXKQCJ2BR43910ZYX',
            'cc8206050e541f79f92', 'FHKKQC963NGSY18ZDZ',
            'R5HKQCIEPOY1DMQOWX', 'ZHMKQC50PFVAPI8PZ6', 'T4UKQCYGECXGVNBWMY',
            'cc82060516c6c141500', 'UPWKQCYVFH7RZOSZ29',
            '2AMKQCE67YOH3TBVYI', '2CUKQCFDVN3EZE2E4C', '7IOKQC1NVGUI1E55CQ',
            'KSTKQC018GNA7HDCAS', 'WIVKQC4Q4FCQJT5M63', 'A4YKQCRYSI5FT5T38',
            'BUNKQCO4HZHZP70F3K', 'YRTKQCNDP343OD5OQJ', '7VEKQCV05EDK0625KI',
            'cc8205d872f532ab14e', 'TPXKQCEGL04KHGMO2X',
            'L9LKQCQ8KJRKHM4D2E', '8QXKQCHJ2EUC7OV8EQ', 'W0XKQCKSFWWJWQ2OSN',
            'I6QKQCFRDTV2LDC8M2', 'XTUKQC7WCIVA5F0NC4', 'F4UKQCPK572VWU2YZQ',
            'JKDKQCF4ND92A088J2', 'COFKQCUXC2H4G3QBYT', 'BNXKQCEBXC1RCOGJNF',
            'Q42KQCKJZGS4IZWHF5', 'P5IKQC88STY3FNTFZ3', '7CXKQC59NSZFXIG1UE',
            'cc8205d87c20350420b', 'FQUKQCWEHOAWUP4QWS',
            '3YTKQCK2W63W0MQBJE', '8HKKQCTEJAOBVH410L', 'HLQKQC0BJIZL0V4EK4',
            'B0UKQC9A54F1GUB7NR', 'EPXKQCQRZP2PNPN7BE',
            'cc82060512042f67e2c', 'XZLKQCRQA9EHPBNZPT',
            'OQXKQC2Y5FVH9PK0JL', 'AXLKQC0YTFAWQ234YD', 'OFXKQC8W0N3N6JP6YQ',
            'MWUKQCD2ZSCECQOCLG', '1ENKQCBPFZTAQJSP4O', 'N7XKQCYD3VSCSZREGJ',
            '2LQKQC62GJUQCJIOK8', 'QXXKQC9PT5FWNT140K', 'VAXKQC19HIFPX61J28',
            '0PXKQCJ9S1M3NNASET', 'K8XKQCDSVLSK422A3K', '52UKQCFYXMFTKIGNBS',
            '7GXKQCMVFU8WR1LKZL', '4UMKQCF07KL2K92CI5', 'LGXKQCJ5OP6MKF9QLN',
            'FZTKQCSTPIQ3C9JC46', 'WMXKQCDUJ4JKQQYCR7', 'R6UKQC939L9FV62UGE',
            'OIUKQCBHUWDGL7DNTI', 'FRTKQC3G6JBJAR2ZPX', 'PIEKQCKUL6OAMS8Q9R',
            'cc8205d887376aacba2', 'LGMKQCQP5M5L18FVTN',
            '8HUKQCRV8B3J2LLQ3B', 'LOUKQC45HUN532HOOM',
            'cc8205d883763f02abd', 'TBXKQC7OHIN28PVCS3',
            ]))

    def test_isbookmarked(self):
        """
        Test IsBookmarked rule.
        """
        rule = IsBookmarked([])
        self.assertEqual(self.filter_with_rule(rule), set([
            '35WJQC1B7T7NPV8OLV', 'AWFKQCJELLUWDY2PD3', 'Q8HKQC3VMRM1M6M7ES',
            ]))

    def test_dupancestor_empty(self):
        """
        Test empty IsDuplicatedAncestorOf rule.
        """
        rule = IsDuplicatedAncestorOf([''])
        self.assertEqual(self.filter_with_rule(rule), set([
            ]))

    def test_dupancestor_nonmatching(self):
        """
        Test non-matching IsDuplicatedAncestorOf rule.
        """
        rule = IsDuplicatedAncestorOf(['I0000'])
        self.assertEqual(self.filter_with_rule(rule), set([
            ]))

    def test_dupancestor_irregular(self):
        """
        Test irregular IsDuplicatedAncestorOf rule.
        """
        rule = IsDuplicatedAncestorOf(['ABCDEFG'])
        self.assertEqual(self.filter_with_rule(rule), set([
            ]))

    def test_dupancestor_matching(self):
        """
        Test matching IsDuplicatedAncestorOf rule.
        """
        rule = IsDuplicatedAncestorOf(['I1631'])
        self.assertEqual(self.filter_with_rule(rule), set([
            'I3VJQCUY5I6UR92507', 'D4VJQC09STQCWD393E',
            ]))

    def test_isrelatedwith_empty(self):
        """
        Test empty IsRelatedWith rule.
        """
        rule = IsRelatedWith([''])
        self.assertEqual(self.filter_with_rule(rule), set([
            ]))

    def test_isrelatedwith_nonmatching(self):
        """
        Test non-matching IsRelatedWith rule.
        """
        rule = IsRelatedWith(['I0000'])
        self.assertEqual(self.filter_with_rule(rule), set([
            'd5839c1237765987724', 'd5839c126d11a754f46',
            ]))

    def test_isrelatedwith_irregular(self):
        """
        Test irregular IsRelatedWith rule.
        """
        rule = IsRelatedWith(['ABCDEFG'])
        self.assertEqual(self.filter_with_rule(rule), set([
            ]))

    def test_isrelatedwith_matching(self):
        """
        Test matching IsRelatedWith rule.
        """
        rule = IsRelatedWith(['I1844'])
        self.assertEqual(self.filter_with_rule(rule), set([
            'HWTKQCSM28EI6WFDHP', 'T4UKQCYGECXGVNBWMY', 'YOTKQCEX2PLG03LZQS',
            'X8UKQCIDY21QIQBDVI', 'F3UKQC7ZV3EYVWTZ8O', '0TTKQCXXY59OCDPLV3',
            'EVTKQCHV2E2PODFD7C', 'BBUKQC5GPRPDJHJAWU', 'FRTKQC3G6JBJAR2ZPX',
            'NDTKQCN95VFLGJ21L', 'SFTKQC26EJ2BYQCRIA', 'MYTKQCVCFOFM32H9GB',
            'B0UKQC9A54F1GUB7NR', 'PTTKQCYN0JR3ZZJNWR', 'F4UKQCPK572VWU2YZQ',
            'LLTKQCX39KCXFSX0U4', 'IXTKQC1BAU1F1WNXKB', '3YTKQCK2W63W0MQBJE',
            'TQTKQCO897BNA1H93B', 'DOTKQCP1MG3VC8D7V2', '3NTKQCZKLMIM6HYFE1',
            'WUTKQCVQCUPFFOGUT8', 'GETKQCPRC2W5YDUYM6', 'YRTKQCNDP343OD5OQJ',
            'U0UKQCBZS0R6WW7LBS', 'J2UKQC897I42M9VHDD', '7MTKQC1QNE4H5RF35S',
            '5FTKQCKT9SDZ8TB03C', 'O1UKQCJD5YHDRW887V', 'EUTKQCFATXRU431YY6',
            'UHTKQCORH3NTZ0FYL3', '2CUKQCFDVN3EZE2E4C', 'RNTKQCMLGRRKQVKDPR',
            'CGTKQC4WO8W3WSQRCX', 'WAUKQCOQ91QCJZWQ9U', 'FZTKQCSTPIQ3C9JC46',
            'AHTKQCM2YFRW3AGSRL', 'WBTKQCC775IAAGIWZD', '8KTKQC407A8CN5O68H',
            '8QTKQCN8ZKY5OWWJZF', 'UKTKQCSL3AUJIWTD2A', 'HAUKQCM3GYGVTREGZS',
            '52UKQCFYXMFTKIGNBS', 'U3UKQCO30PWAK6JQBA', 'R6UKQC939L9FV62UGE',
            'TZTKQCR39A060AQ63C', 'X9UKQCFELSDAQ2TDP1', 'B1UKQCBR49WB3134PN',
            'KSTKQC018GNA7HDCAS', 'FJTKQCJCMAHJOA9NHI', 'HITKQCWJSCZX2AN6NP',
            'WVTKQCZC91I63LHEE7', '0DTKQC6KBOS69LQJ35',
            ]))

    def test_hasidof_empty(self):
        """
        Test empty HasIdOf rule.
        """
        rule = HasIdOf([''])
        self.assertEqual(self.filter_with_rule(rule), set([
            ]))

    def test_hasidof_nonmatching(self):
        """
        Test non-matching HasIdOf rule.
        """
        rule = HasIdOf(['I0000'])
        self.assertEqual(self.filter_with_rule(rule), set([
            'd5839c1237765987724'
            ]))

    def test_hasidof_irregular(self):
        """
        Test irregular HasIdOf rule.
        """
        rule = HasIdOf(['ABCDEFG'])
        self.assertEqual(self.filter_with_rule(rule), set([
            ]))

    def test_hasidof_matching(self):
        """
        Test matching HasIdOf rule.
        """
        rule = HasIdOf(['I0044'])
        self.assertEqual(self.filter_with_rule(rule), set([
            'GNUJQCL9MD64AM56OH',
            ]))

    def test_isdefaultperson(self):
        """
        Test IsDefaultPerson rule.
        """
        rule = IsDefaultPerson([])
        self.assertEqual(self.filter_with_rule(rule), set([
            'GNUJQCL9MD64AM56OH',
            ]))

    def test_isfemale(self):
        """
        Test IsFemale rule.
        """
        rule = IsFemale([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 940)

    def test_ismale(self):
        """
        Test IsMale rule.
        """
        rule = IsMale([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 1168)

    def test_missingparent(self):
        """
        Test MissingParent rule.
        """
        rule = MissingParent([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 868)

    def test_multiplemarriages(self):
        """
        Test MultipleMarriages rule.
        """
        rule = MultipleMarriages([])
        self.assertEqual(self.filter_with_rule(rule), set([
            'R1VKQCJWNP24VN7BO', 'ZTVJQCTSMI85EGMXFM', 'ENTJQCZXQV1IRKJXUL',
            '44WJQCLCQIPZUB0UH', 'SMWJQCXQ6I2GEXSPK9', 'DN3KQC1URTED410L3R',
            '5FYJQC86G8EZ0L4E4B', '5F4KQCJRU8ZKL6SILT', '0YNKQC5U4EQGVNUZD8',
            'YRYJQCE3RF4U8A59UB', 'APWKQCI6YXAXBLC33I', 'XSKKQC6GGKLAYANWAF',
            '0FQKQCOQD0VRVJPTSD', 'B3UJQCZHDXII99AWW4',
            'cc8205d872f532ab14e', 'SS1KQCWWF9488Q330U',
            'OCYJQCS8YT7JO8KIMO', 'I6HKQCQF72V2N56JQ5', '6YWJQC86FBVN0J6JS',
            'KYNKQCVA6FE65ONFIQ', 'SHAKQCNY5IXO30GUAB', 'O5XKQC3V6BPJI13J24',
            'ZN7KQC3RLB82EXF1QF', 'CIYJQCF3UK12DL0S2Y', 'H3XJQCFJ4FP4U2WGZC',
            'cc82060504445ab6deb', '4E4KQC1K4XUEX29IJO',
            '0XVJQCJUNJY40WDSMA', '1WUJQCHNH76G6YD3A', 'IH3KQCM1VZPRKLBLK7',
            '242KQCBALBOD8ZK5VI', '8G4KQCS6C1AOM6ZGR3', 'I1EKQCGGDSUD8ILUW4',
            'X8BKQCSFF4AET5MY23', 'RJWJQCN1XKXRN5KMCP', 'ZWNKQC9DAZ3C6UHUAV',
            '9QUJQCCSWRZNSAPCR', 'HI0KQCG9TGT5AAIPU', 'DI4KQC3S1AO27VWOLN',
            'QBDKQCH2IU6N8IXMFE', 'DK2KQCJYW14VXUJ85', '117KQCBB32RMTTV4G6',
            '0QLKQCFTQMNVGCV4GM', 'D2OKQCGDNPT3BH4WH', 'CAYJQCKOL49OF7XWB3',
            'ZQGKQCGHS67Q4IMHEG', 'OEXJQCQJHF2BLSAAIS', 'UKYJQC70LIZQ11BP89',
            'FF2KQCRBSPCG1QY97', 'L6EKQCO8QYL2UO2MQO',
            ]))

    def test_nevermarried(self):
        """
        Test NeverMarried rule.
        """
        rule = NeverMarried([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 751)

    def test_nobirthdate(self):
        """
        Test NoBirthdate rule.
        """
        rule = NoBirthdate([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 981)

    def test_nodeathdate(self):
        """
        Test NoDeathdate rule.
        """
        rule = NoDeathdate([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 1603)

    def test_peopleprivate(self):
        """
        Test PeoplePrivate rule.
        """
        # TODO: example.gramps has no people marked private
        rule = PeoplePrivate([])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_peoplepublic(self):
        """
        Test PeoplePublic rule.
        """
        rule = PeoplePublic([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 2128)

    def test_personwithincompleteevent(self):
        """
        Test PersonWithIncompleteEvent rule.
        """
        rule = PersonWithIncompleteEvent([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 745)

    def test_relationshipbookmarks(self):
        """
        Test RelationshipPathBetweenBookmarks rule.
        """
        rule = RelationshipPathBetweenBookmarks([])
        self.assertEqual(self.filter_with_rule(rule), set([
            '44WJQCLCQIPZUB0UH', '35WJQC1B7T7NPV8OLV', 'AWFKQCJELLUWDY2PD3',
            'D3WJQCCGV58IP8PNHZ', 'Q8HKQC3VMRM1M6M7ES',
            ]))

    def test_hassoundexname(self):
        """
        Test HasSoundexName rule.
        """
        rule = HasSoundexName(['garner'])
        self.assertEqual(len(self.filter_with_rule(rule)), 73)

    def test_hasnameof(self):
        """
        Test HasNameOf rule.
        """
        rule = HasNameOf(['Lewis', 'Garner', 'Dr.', 'Sr', 'Anderson',
                          'Big Louie', 'von', 'Zieliński', None, None, None])
        self.assertEqual(self.filter_with_rule(rule), set([
            'GNUJQCL9MD64AM56OH']))


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2016 Tom Samstag
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Unittest that tests family-specific filter rules
"""
import unittest
import os

from ....db.utils import import_as_dict
from ....filters import GenericFilterFactory
from ....const import DATA_DIR
from ....user import User

from ..family import (
    AllFamilies, HasRelType, HasGallery, HasIdOf, HasLDS, HasNote, RegExpIdOf,
    HasNoteRegexp, HasReferenceCountOf, HasSourceCount, HasSourceOf,
    HasCitation, FamilyPrivate, HasEvent, HasAttribute, IsBookmarked,
    MatchesSourceConfidence, FatherHasNameOf, FatherHasIdOf, MotherHasNameOf,
    MotherHasIdOf, ChildHasNameOf, ChildHasIdOf, ChangedSince, HasTag,
    HasTwins, IsAncestorOf, IsDescendantOf)

TEST_DIR = os.path.abspath(os.path.join(DATA_DIR, "tests"))
EXAMPLE = os.path.join(TEST_DIR, "example.gramps")
GenericFamilyFilter = GenericFilterFactory('Family')

class BaseTest(unittest.TestCase):
    """
    Family rule tests.
    """

    @classmethod
    def setUpClass(cls):
        """
        Import example database.
        """
        cls.db = import_as_dict(EXAMPLE, User())

    def filter_with_rule(self, rule):
        """
        Apply a filter with the given rule.
        """
        filter_ = GenericFamilyFilter()
        filter_.add_rule(rule)
        results = filter_.apply(self.db)
        return set(results)

    def test_allfamilies(self):
        """
        Test AllFamilies rule.
        """
        rule = AllFamilies([])
        self.assertEqual(len(self.filter_with_rule(rule)),
                         self.db.get_number_of_families())

    def test_hasreltype(self):
        """
        Test HasRelType rule.
        """
        rule = HasRelType(['Married'])
        self.assertEqual(len(self.filter_with_rule(rule)), 750)

    def test_hasgallery(self):
        """
        Test HasGallery rule.
        """
        rule = HasGallery(['0', 'greater than'])
        self.assertEqual(self.filter_with_rule(rule),
                         set(['9OUJQCBOHW9UEK9CNV']))

    def test_hasidof(self):
        """
        Test HasIdOf rule.
        """
        rule = HasIdOf(['F0001'])
        self.assertEqual(self.filter_with_rule(rule),
                         set(['48TJQCGNNIR5SJRCAK']))

    def test_haslds(self):
        """
        Test HasLDS rule.
        """
        rule = HasLDS(['0', 'greater than'])
        self.assertEqual(self.filter_with_rule(rule),
                         set(['9OUJQCBOHW9UEK9CNV']))

    def test_hasnote(self):
        """
        Test HasNote rule.
        """
        rule = HasNote([])
        self.assertEqual(self.filter_with_rule(rule),
                         set(['9OUJQCBOHW9UEK9CNV']))

    def test_regexpidof(self):
        """
        Test RegExpIdOf rule.
        """
        rule = RegExpIdOf(['F000.'], use_regex=True)
        self.assertEqual(self.filter_with_rule(rule), set([
            'LOTJQC78O5B4WQGJRP', 'UPTJQC4VPCABZUDB75',
            'NBTJQCIX49EKOCIHBP', 'C9UJQCF6ETBTV2MRRV',
            '74UJQCKV8R4NBNHCB', '4BTJQCL4CHNA5OUTKF',
            '48TJQCGNNIR5SJRCAK', '4YTJQCTEH7PQUU4AD',
            'MTTJQC05LKVFFLN01A', 'd5839c123c034ef82ab',
            ]))

    def test_hasnoteregexp(self):
        """
        Test HasNoteRegexp rule.
        """
        rule = HasNoteRegexp(['.'], use_regex=True)
        self.assertEqual(self.filter_with_rule(rule),
                         set(['9OUJQCBOHW9UEK9CNV']))

    def test_hasreferencecountof(self):
        """
        Test HasReferenceCountOf rule.
        """
        rule = HasReferenceCountOf(['greater than', '12'])
        self.assertEqual(self.filter_with_rule(rule), set([
            '29IKQCMUNFTIBV653N', '8OUJQCUVZ0XML7BQLF', 'UPTJQC4VPCABZUDB75',
            '9NWJQCJGLXUR3AQSFJ', '5G2KQCGBTS86UVSRG5', 'WG2KQCSY9LEFDFQHMN',
            'MTTJQC05LKVFFLN01A', 'C2VJQC71TNHO7RBBMX', 'QIDKQCJQ37SIUQ3UFU',
            'DV4KQCX9OBVQ74H77F']))

    def test_hassourcecount(self):
        """
        Test HasSourceCount rule.
        """
        rule = HasSourceCount(['1', 'greater than'])
        self.assertEqual(self.filter_with_rule(rule),
                         set(['9OUJQCBOHW9UEK9CNV']))

    def test_hassourceof(self):
        """
        Test HasSourceOf rule.
        """
        rule = HasSourceOf(['S0001'])
        self.assertEqual(self.filter_with_rule(rule),
                         set(['9OUJQCBOHW9UEK9CNV']))

    def test_hascitation(self):
        """
        Test HasCitation rule.
        """
        rule = HasCitation(['page 10', '', '2'])
        self.assertEqual(self.filter_with_rule(rule),
                         set(['9OUJQCBOHW9UEK9CNV']))

    def test_familyprivate(self):
        """
        Test FamilyPrivate rule.
        """
        rule = FamilyPrivate([])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_hasevent(self):
        """
        Test HasEvent rule.
        """
        rule = HasEvent(['Marriage', 'before 1900', 'USA', '', 'Garner'])
        self.assertEqual(self.filter_with_rule(rule), set([
            'KSFKQCP4V0YXGM1LR9', '8ZFKQC3FRSHACOJBOU', '3XFKQCE7QUDJ99AVNV',
            'OVFKQC51DX0OQUV3JB', '9OUJQCBOHW9UEK9CNV']))

    def test_hasattribute(self):
        """
        Test HasAttribute rule.
        """
        rule = HasAttribute(['Number of Children', ''])
        self.assertEqual(self.filter_with_rule(rule),
                         set(['9OUJQCBOHW9UEK9CNV']))

    def test_isbookmarked(self):
        """
        Test IsBookmarked rule.
        """
        rule = IsBookmarked([])
        self.assertEqual(self.filter_with_rule(rule),
                         set(['9OUJQCBOHW9UEK9CNV']))

    def test_matchessourceconfidence(self):
        """
        Test MatchesSourceConfidence rule.
        """
        rule = MatchesSourceConfidence(['0'])
        self.assertEqual(len(self.filter_with_rule(rule)), 734)

    def test_fatherhasnameof(self):
        """
        Test FatherHasNameOf rule.
        """
        rule = FatherHasNameOf(['', '', 'Dr.', '', '', '', '', '', '', '',
                                ''])
        self.assertEqual(self.filter_with_rule(rule),
                         set(['9OUJQCBOHW9UEK9CNV']))

    def test_fatherhasidof(self):
        """
        Test FatherHasIdOf rule.
        """
        rule = FatherHasIdOf(['I0106'])
        self.assertEqual(self.filter_with_rule(rule),
                         set(['8OUJQCUVZ0XML7BQLF']))

    def test_motherhasnameof(self):
        """
        Test MotherHasNameOf rule.
        """
        rule = MotherHasNameOf(['', 'Alvarado', '', '', '', '', '', '', '', '',
                                ''])
        self.assertEqual(self.filter_with_rule(rule), set([
            'EM3KQC48HFLA02TF8D', 'K9NKQCBG105ECXZ48D',
            '2QMKQC5YWNAWZMG6VO', '6JUJQCCAXGENRX990K']))

    def test_motherhasidof(self):
        """
        Test MotherHasIdOf rule.
        """
        rule = MotherHasIdOf(['I0107'])
        self.assertEqual(self.filter_with_rule(rule),
                         set(['8OUJQCUVZ0XML7BQLF']))

    def test_childhasnameof(self):
        """
        Test ChildHasNameOf rule.
        """
        rule = ChildHasNameOf(['Eugene', '', '', '', '', '', '', '', '', '',
                               ''])
        self.assertEqual(self.filter_with_rule(rule), set([
            'D1YJQCGLEIBPPLNL4B', '5GTJQCXVYVAIQTBVKA', 'I42KQCM3S926FMJ91O',
            '7CTJQCFJVBQSY076A6', '9OUJQCBOHW9UEK9CNV', '9IXJQCX18AHUFPQHEZ',
            '9NWJQCJGLXUR3AQSFJ']))

    def test_childhasidof(self):
        """
        Test ChildHasIdOf rule.
        """
        rule = ChildHasIdOf(['I0001'])
        self.assertEqual(self.filter_with_rule(rule),
                         set(['48TJQCGNNIR5SJRCAK']))

    def test_changedsince(self):
        """
        Test ChangedSince rule.
        """
        rule = ChangedSince(['2008-01-01', '2014-01-01'])
        self.assertEqual(self.filter_with_rule(rule),
                         set(['9OUJQCBOHW9UEK9CNV']))

    def test_hastag(self):
        """
        Test HasTag rule.
        """
        rule = HasTag(['ToDo'])
        self.assertEqual(self.filter_with_rule(rule),
                         set(['9OUJQCBOHW9UEK9CNV']))

    def test_hastwins(self):
        """
        Test HasTwins rule.
        """
        rule = HasTwins([])
        self.assertEqual(self.filter_with_rule(rule), set([
            'SD6KQC7LB8MYGA7F5W', '8OUJQCUVZ0XML7BQLF', '1BVJQCNTFAGS8273LJ',
            '5IUJQCRJY47YQ8PU7N', 'ZLUJQCPDV93OR8KHB7', '4U2KQCBXG2VTPH6U1F',
            ]))

    def test_isancestorof(self):
        """
        Test IsAncestorOf rule.
        """
        rule = IsAncestorOf(['F0031', '0'])
        self.assertEqual(self.filter_with_rule(rule), set([
            '4AXJQC96KTN3WGPTVE', '1RUJQCYX9QL1V45YLD', '5GTJQCXVYVAIQTBVKA',
            'X3WJQCSF48F6809142', 'NSVJQC89IHEEBIPDP2', '9OUJQCBOHW9UEK9CNV',
            '1RUJQCCL9MVRYLMTBO', 'RRVJQC5A8DDHQFPRDL', '0SUJQCOS78AXGWP8QR',
            '57WJQCTBJKR5QYPS6K', '8OUJQCUVZ0XML7BQLF', '7PUJQC4PPS4EDIVMYE'
            ]))

    def test_isdescendantof(self):
        """
        Test IsDescendantOf rule.
        """
        rule = IsDescendantOf(['F0031', '0'])
        self.assertEqual(self.filter_with_rule(rule), set([
            'SFXJQCLE8PIG7PH38J', 'UCXJQCC5HS8VXDKWBM', 'IIEKQCRX89WYBHKB7R',
            'XDXJQCMWU5EIV8XCRF', '7BXJQCU22OCA4HN38A', '3FXJQCR749H2H7G321',
            'IEXJQCFUN95VENI6BO', '4FXJQC7656WDQ3HJGW', 'FLEKQCRVG3O1UA9YUB',
            'BCXJQC9AQ0DBXCVLEQ', '9SEKQCAAWRUCIO7A0M', 'DDXJQCVT5X72TOXP0C',
            'CGXJQC515QL9RLPQTU', 'XGXJQCNVZH2PWRMVAH', 'RBXJQCUYMQR2KRMDFY'
            ]))


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2015-      Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# https://en.wikipedia.org/wiki/Miscellaneous_Symbols
# http://www.w3schools.com/charsets/ref_utf_symbols.asp
#

#-------------------------------------------------------------------------
#
# Standard python modules
#
#-------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
_ = glocale.translation.sgettext

# pylint: disable=superfluous-parens
# pylint: disable=anomalous-unicode-escape-in-string

class Symbols(object):
    # genealogical symbols
    SYMBOL_FEMALE                    = 0
    SYMBOL_MALE                      = 1
    SYMBOL_ASEXUAL_SEXLESS           = 2 # Unknown
    SYMBOL_LESBIAN                   = 3
    SYMBOL_MALE_HOMOSEXUAL           = 4
    SYMBOL_HETEROSEXUAL              = 5
    SYMBOL_HERMAPHRODITE             = 6
    SYMBOL_TRANSGENDER               = 7
    SYMBOL_NEUTER                    = 8

    SYMBOL_ILLEGITIM                 = 9
    SYMBOL_BIRTH                     = 10
    SYMBOL_BAPTISM                   = 11 # CHRISTENING
    SYMBOL_ENGAGED                   = 12
    SYMBOL_MARRIAGE                  = 13
    SYMBOL_DIVORCE                   = 14
    SYMBOL_UNMARRIED_PARTNERSHIP     = 15
    SYMBOL_BURIED                    = 16
    SYMBOL_CREMATED                  = 17 # Funeral urn
    SYMBOL_KILLED_IN_ACTION          = 18
    SYMBOL_EXTINCT                   = 19

    all_symbols = [
               # Name                     UNICODE       SUBSTITUTION
               (_("Female"),              '\u2640',     ""),
               (_("Male"),                '\u2642',     ""),
               (_("Asexuality, sexless, genderless"), '\u26aa', ""),
               (_("Lesbianism"),          '\u26a2',     "oo"),
               (_("Male homosexuality"),  '\u26a3',     "oo"),
               (_("Heterosexuality"),     '\u26a4',     "oo"),
               (_("Transgender, hermaphrodite (in entomology)"), '\u26a5', ""),
               (_("Transgender"),         '\u26a6',     ""),
               (_("Neuter"),              '\u26b2',     ""),

               (_("Illegitimate"),        '\u229b',     ""),
               (_("Birth"),               '\u002a',     "*"),
               (_("Baptism/Christening"), '\u007e',     "~"),
               (_("Engaged"),             '\u26ac',     "o"),
               (_("Marriage"),            '\u26ad',     "oo"),
               (_("Divorce"),             '\u26ae',     "o|o"),
               (_("Unmarried partnership"), '\u26af',   "o-o"),
               (_("Buried"),              '\u26b0',     "d"),
               (_("Cremated/Funeral urn"), '\u26b1',    "d"),
               (_("Killed in action"),    '\u2694',     "d"),
               (_("Extinct"),             '\u2021',     ""),
              ]

    # genealogical death symbols
    DEATH_SYMBOL_NONE                      = 0
    DEATH_SYMBOL_X                         = 1
    DEATH_SYMBOL_SKULL                     = 2
    DEATH_SYMBOL_ANKH                      = 3
    DEATH_SYMBOL_ORTHODOX_CROSS            = 4
    DEATH_SYMBOL_CHI_RHO                   = 5
    DEATH_SYMBOL_LORRAINE_CROSS            = 6
    DEATH_SYMBOL_JERUSALEM_CROSS           = 7
    DEATH_SYMBOL_STAR_CRESCENT             = 8
    DEATH_SYMBOL_WEST_SYRIAC_CROSS         = 9
    DEATH_SYMBOL_EAST_SYRIAC_CROSS         = 10
    DEATH_SYMBOL_HEAVY_GREEK_CROSS         = 11
    DEATH_SYMBOL_LATIN_CROSS               = 12
    DEATH_SYMBOL_SHADOWED_LATIN_CROSS      = 13
    DEATH_SYMBOL_MALTESE_CROSS             = 14
    DEATH_SYMBOL_STAR_OF_DAVID             = 15
    DEATH_SYMBOL_DEAD                      = 16

    # The following is used in the global preferences in the display tab.
    #                Name
    #                                                  UNICODE    SUBSTITUTION
    death_symbols = [(_("Nothing"),                    "",        ""),
                     ("x",                             "x",       "x"),
                     (_("Skull and crossbones"),       "\u2620",  "+"),
                     (_("Ankh"),                       "\u2625",  "+"),
                     (_("Orthodox cross"),             "\u2626",  "+"),
                     (_("Chi rho"),                    "\u2627",  "+"),
                     (_("Cross of Lorraine"),          "\u2628",  "+"),
                     (_("Cross of Jerusalem"),         "\u2629",  "+"),
                     (_("Star and crescent"),          "\u262a",  "+"),
                     (_("West Syriac cross"),          "\u2670",  "+"),
                     (_("East Syriac cross"),          "\u2671",  "+"),
                     (_("Heavy Greek cross"),          "\u271a",  "+"),
                     (_("Latin cross"),                "\u271d",  "+"),
                     (_("Shadowed White Latin cross"), "\u271e",  "+"),
                     (_("Maltese cross"),              "\u2720",  "+"),
                     (_("Star of David"),              "\u2721",  "+"),
                     (_("Dead"),                       _("Dead"), _("Dead"))
                ]

    def __init__(self):
        self.symbols = None
    #
    # functions for general symbols
    #
    def get_symbol_for_html(self, symbol):
        """ return the html string like '&#9898;' """
        return '&#%d;' % ord(self.all_symbols[symbol][1])

    def get_symbol_name(self, symbol):
        """
        Return the name of the symbol.
        """
        return self.all_symbols[symbol][0]

    def get_symbol_for_string(self, symbol):
        """ return the utf-8 character like '\u2670' """
        return self.all_symbols[symbol][1]

    def get_symbol_fallback(self, symbol):
        """
        Return the replacement string.
        This is used if the utf-8 symbol in not present within a font.
        """
        return self.all_symbols[symbol][2]

    #
    # functions for death symbols
    #
    def get_death_symbols(self):
        """
        Return the list of death symbols.
        This is used in the global preference to choose which symbol we'll use.
        """
        return self.death_symbols

    def get_death_symbol_name(self, symbol):
        """
        Return the name of the symbol.
        """
        return self.death_symbols[symbol][0]

    def get_death_symbol_for_html(self, symbol):
        """
        return the html string like '&#9898;'.
        """
        return '&#%d;' % ord(self.death_symbols[symbol][1])

    def get_death_symbol_for_char(self, symbol):
        """
        Return the utf-8 character for the symbol.
        """
        return self.death_symbols[symbol][1]

    def get_death_symbol_fallback(self, symbol):
        """
        Return the string replacement for the symbol.
        """
        return self.death_symbols[symbol][2]

    #
    # functions for all symbols
    #
    def get_how_many_symbols(self):
        return len(self.death_symbols) + len(self.all_symbols) - 4



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2005  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Rewritten in 2008 for 3.x version by Łukasz Rymarczyk
# Written in 2007 by Piotr Czubaszek, largely based on rel_de.py by Alex Roitman.

# PL: Po objaśnienia oznaczania relacji zobacz relationship.py
# EN: For more information see relationship.py
#
"""
Polish-specific definitions of relationships.
"""

#-------------------------------------------------------------------------
#
# Gramps modules
#
#-------------------------------------------------------------------------

from gramps.gen.lib import Person
import gramps.gen.relationship

#-------------------------------------------------------------------------


# określa liczebnik porządkowy

_level_name = [ "pierwszego", "drugiego", "trzeciego", "czwartego", "piątego",
    "szóstego", "siódmego", "ósmego", "dziewiątego", "dziesiątego",
    "jedenastego", "dwunastego","trzynastego", "czternastego", "piętnastego",
    "szesnastego", "siedemnastego", "osiemnastego","dziewiętnastego", "dwudziestego", ]


_father_level = [ "",
  "ojciec",
  "dziadek",
  "pradziadek",
  "prapradziadek",
  "praprapradziadek",
  "prapraprapradziadek",
  "praprapraprapradziadek",
  "prapraprapraprapradziadek",
  "praprapraprapraprapradziadek",
  "prapraprapraprapraprapradziadek",
]


_mother_level = [ "",
  "matka",
  "babcia",
  "prababcia",
  "praprababcia",
  "prapraprababcia",
  "praprapraprababcia",
  "prapraprapraprababcia",
  "praprapraprapraprababcia",
  "prapraprapraprapraprababcia",
  "praprapraprapraprapraprababcia",
]

_son_level = [ "",
  "syn",
  "wnuk",
  "prawnuk",
  "praprawnuk",
  "prapraprauwnuk",
  "praprapraprauwnuk",
  "prapraprapraprawnuk",
  "praprapraprapraprawnuk",
  "prapraprapraprapraprawnuk",
  "praprapraprapraprapraprawnuk",
]

_daughter_level = [ "",
  "córka",
  "wnuczka",
  "prawnuczka",
  "praprawnuczka",
  "prapraprauwnuczka",
  "praprapraprauwnuczka",
  "prapraprapraprawnuczka",
  "praprapraprapraprawnuczka",
  "prapraprapraprapraprawnuczka",
  "praprapraprapraprapraprawnuczka",
]

_sister_level_of_male = [ "", "siostra", "ciotka stryjeczna",
  "babcia stryjeczna",
  "prababcia stryjeczna",
  "praprababcia stryjeczna",
  "prapraprababcia stryjeczna",
  "praprapraprababcia stryjeczna",
  "prapraprapraprababcia stryjeczna",
  "praprapraprapraprababcia stryjeczna",
  "prapraprapraprapraprababcia stryjeczna",
  "praprapraprapraprapraprababcia stryjeczna",
]

_sister_level_of_female = [ "", "siostra", "ciotka",
  "babcia cioteczna",
  "prababcia cioteczna",
  "praprababcia cioteczna",
  "prapraprababcia cioteczna",
  "praprapraprababcia cioteczna",
  "prapraprapraprababcia cioteczna",
  "praprapraprapraprababcia cioteczna",
  "prapraprapraprapraprababcia cioteczna",
  "praprapraprapraprapraprababcia cioteczna",
]

_brother_level_of_male = [ "", "brat", "stryj",
  "dziadek stryjeczny",
  "pradziadek stryjeczny",
  "prapradziadek stryjeczny",
  "praprapradziadek stryjeczny",
  "prapraprapradziadek stryjeczny",
  "praprapraprapradziadek stryjeczny",
  "prapraprapraprapradziadek stryjeczny",
  "praprapraprapraprapradziadek stryjeczny",
  "prapraprapraprapraprapradziadek stryjeczny",
]

_brother_level_of_female = [ "", "brat", "wuj",
  "dziadek cioteczny",
  "pradziadek cioteczny",
  "prapradziadek cioteczny",
  "praprapradziadek cioteczny",
  "prapraprapradziadek cioteczny",
  "praprapraprapradziadek cioteczny",
  "prapraprapraprapradziadek cioteczny",
  "praprapraprapraprapradziadek cioteczny",
  "prapraprapraprapraprapradziadek cioteczny",
]

_nephew_level_of_brothers_son = [ "", "bratanek",
  "syn bratanka",
  "wnuk bratanka",
  "prawnuk bratanka",
  "praprawnuk bratanka",
  "prapraprawnuk bratanka",
  "praprapraprawnuk bratanka",
  "prapraprapraprawnuk bratanka",
  "praprapraprapraprawnuk bratanka",
  "prapraprapraprapraprawnuk bratanka",
]

_nephew_level_of_brothers_daughter = [ "", "bratanica",
  "syn bratanicy",
  "wnuk bratanicy",
  "prawnuk bratanicy",
  "praprawnuk bratanicy",
  "prapraprawnuk bratanicy",
  "praprapraprawnuk bratanicy",
  "prapraprapraprawnuk bratanicy",
  "praprapraprapraprawnuk bratanicy",
  "prapraprapraprapraprawnuk bratanicy",
  "praprapraprapraprapraprawnuk bratanicy",
]

_nephew_level_of_sisters_son = [ "", "siostrzeniec",
  "syn siostrzeńca",
  "wnuk siostrzeńca",
  "prawnuk siostrzeńca",
  "praprawnuk siostrzeńca",
  "prapraprawnuk siostrzeńca",
  "praprapraprawnuk siostrzeńca",
  "prapraprapraprawnuk siostrzeńca",
  "praprapraprapraprawnuk siostrzeńca",
  "prapraprapraprapraprawnuk siostrzeńca",
]

_nephew_level_of_sisters_daughter = [ "", "siostrzenica",
  "syn siostrzenicy",
  "wnuk siostrzenicy",
  "prawnuk siostrzenicy",
  "praprawnuk siostrzenicy",
  "prapraprawnuk siostrzenicy",
  "praprapraprawnuk siostrzenicy",
  "prapraprapraprawnuk siostrzenicy",
  "praprapraprapraprawnuk siostrzenicy",
  "prapraprapraprapraprawnuk siostrzenicy",
]

_niece_level_of_brothers_son = [ "", "bratanica",
  "córka bratanka",
  "wnuczka bratanka",
  "prawnuczka bratanka",
  "praprawnuczka bratanka",
  "prapraprawnuczka bratanka",
  "praprapraprawnuczka bratanka",
  "prapraprapraprawnuczka bratanka",
  "praprapraprapraprawnuczka bratanka",
]

_niece_level_of_brothers_daughter = [ "", "bratanica",
  "córka bratanicy",
  "wnuczka bratanicy",
  "prawnuczka bratanicy",
  "praprawnuczka bratanicy",
  "prapraprawnuczka bratanicy",
  "praprapraprawnuczka bratanicy",
  "prapraprapraprawnuczka bratanicy",
  "praprapraprapraprawnuczka bratanicy",
]

_niece_level_of_sisters_son = [ "", "siostrzenica",
  "córka siostrzeńca",
  "wnuczka siostrzeńca",
  "prawnuczka siostrzeńca",
  "praprawnuczka siostrzeńca",
  "prapraprawnuczka siostrzeńca",
  "praprapraprawnuczka siostrzeńca",
  "prapraprapraprawnuczka siostrzeńca",
  "praprapraprapraprawnuczka siostrzeńca",
]

_niece_level_of_sisters_daughter = [ "", "siostrzenica",
  "córka siostrzenicy",
  "wnuczka siostrzenicy",
  "prawnuczka siostrzenicy",
  "praprawnuczka siostrzenicy",
  "prapraprawnuczka siostrzenicy",
  "praprapraprawnuczka siostrzenicy",
  "prapraprapraprawnuczka siostrzenicy",
  "praprapraprapraprawnuczka siostrzenicy",
]

#-------------------------------------------------------------------------
#
#
#
#-------------------------------------------------------------------------

class RelationshipCalculator(gramps.gen.relationship.RelationshipCalculator):
    """
    RelationshipCalculator Class
    """

    def __init__(self):
        gramps.gen.relationship.RelationshipCalculator.__init__(self)

    def get_son(self, level, inlaw=''):
        """
        Podaje tekst zawierający informację, jak bardzo potomek męski
        (np. syn) jest spokrewniony do danej osoby
        """

        # Określ, czy osoba jest przybraną, czy rodzoną
        if inlaw == '':
            t_inlaw = ""
        else:
            t_inlaw = "przybrany "

        # TODO: dodać rozpoznawanie pasierb/pasierbica
        if level >= 0 and level < len(_son_level):
            return t_inlaw +_son_level[level]

        elif level >= len(_son_level) and (level - 1) < len(_level_name):
            return t_inlaw + \
                "potomek męski %s pokolenia" % _level_name[level - 1]

        else:
            return t_inlaw + \
                "potomek męski w %d pokoleniu" % level



    def get_daughter(self, level, inlaw=''):
        """
        Podaje tekst zawierający informację, jak bardzo potomek żeński
        (np. córka) jest spokrewniony do danej osoby
        """

        # Określ, czy osoba jest przybraną, czy rodzoną
        #   + stwórz obie formy (męską i żeńską)
        if inlaw == '':
            t_inlaw = ""
            t_inlawM = ""
        else:
            t_inlaw = "przybrana "
            t_inlawM = "przybrany "

        # TODO: dodać rozpoznawanie pasierb/pasierbica
        if level >= 0 and level < len(_daughter_level):
            return t_inlaw + _daughter_level[level]

        elif level >= len(_daughter_level) and (level - 1) < len(_level_name):
            return t_inlawM + \
                    "potomek żeński %s pokolenia" % _level_name[level - 1]
        else:
            return t_inlawM + \
                    "potomek żeński w %d pokoleniu" % level


    def get_child_unknown(self, level, inlaw=''):
        """
        Podaje tekst zawierający informację, jak bardzo potomek
        o nieokreślonej płci jest spokrewniony dodanej osoby
        """

        # Określ, czy osoba jest przybraną, czy rodzoną
        if inlaw == '':
            t_inlaw = ""
        else:
            t_inlaw = "przybrany "

        if level == 1:
            if inlaw == '' :
                return "dziecko"
            else:
                return "przybrane dziecko"

        elif level >= 1 and  (level - 1) < len(_level_name):
            return t_inlaw + "potomek %s pokolenia" %  _level_name[level - 1]

        else:
            return t_inlaw + "potomek w %d pokoleniu" % level


    def get_sword_distaff(self, level, reltocommon, spacebefore = ""):
        """
        PL: Generuje relację po mieczu/po kądzieli
        EN: Generate relation 'by sword' or 'by distaff', polish specific
        """
        if level <= 1:
            return ""

        elif level == 2:
            # dziadek/babcia

            if reltocommon[0] == self.REL_FATHER:
                # ze strony rodzonego ojca
                return spacebefore + "po mieczu"
            elif reltocommon[0] == self.REL_MOTHER:
                # ze strony rodzonej matki
                return spacebefore + "po kądzieli"
            else:
                # relacja inna niż rodzona
                return ""

        elif level == 3:
            # pradziadek/prababcia

            if (reltocommon[0] == self.REL_FATHER) \
                & (reltocommon[1] == self.REL_FATHER):
                # pradziadek od dziadka ze strony ojca
                return spacebefore + "podwójnego miecza"

            elif (reltocommon[0] == self.REL_FATHER) \
                & (reltocommon[1] == self.REL_MOTHER):
                # pradziadek od babci ze strony ojca
                return spacebefore + "raz po mieczu, dalej po kądzieli"

            elif (reltocommon[0] == self.REL_MOTHER) \
                & (reltocommon[1] == self.REL_FATHER):
                # pradziadek od dziadka ze strony matki
                return spacebefore + "raz po kądzieli, dalej po mieczu"

            elif (reltocommon[0] == self.REL_MOTHER) \
                & (reltocommon[1] == self.REL_MOTHER):
                # pradziadek od babci ze strony matki
                return spacebefore + "podwójnej kądzieli"

            else:
                # relacja inna niż rodzona
                return ""

        elif level == 4:
            # prapradziadek/praprababcia

            if (reltocommon[0] == self.REL_FATHER) \
                & (reltocommon[1] == self.REL_FATHER) \
                & (reltocommon[2] == self.REL_FATHER):
                # tzw. linia męska
                return spacebefore + "potrójnego miecza"

            if (reltocommon[0] == self.REL_FATHER) \
                & (reltocommon[1] == self.REL_FATHER) \
                & (reltocommon[2] == self.REL_FATHER):
                # tzw. linia żeńska
                return spacebefore + "potrójnego miecza"

            else:
                return ""

        else:
            return ""



    def get_father(self, level, reltocommon, inlaw=''):
        """
        Podaje tekst zawierający informację, jak bardzo przodek męski
        (np. ojciec) jest spokrewniony do danej osoby
        """
        if inlaw == '':
            t_inlaw = ""
        else:
            t_inlaw = "przybrany "

        if level >= 0 and level < len(_father_level):
            # Jeśli znasz bezpośrednią nazwę relacji, to ją zastosuj

            if level == 1:
                # ojciec
                return t_inlaw + _father_level[level]

            elif (level >= 2) & (level <= 4):
                # dziadek, pradziadek, prapradziadek
                return t_inlaw + _father_level[level] \
                    + self.get_sword_distaff(level, reltocommon, ' ')

            else:
                return t_inlaw + _father_level[level]

        elif level >= len(_father_level) and (level - 1) < len(_level_name):
            # jeśli istnieje liczebnik dla danej liczby
            return t_inlaw + \
                    "przodek męski %s pokolenia" % (_level_name[level - 1])

        else:
            # dla pozostałych przypadków wypisz relację liczbowo
            return t_inlaw + \
                    "przodek męski w %d pokoleniu" % level


    def get_mother(self, level, reltocommon, inlaw=''):
        """
        Podaje tekst zawierający informację, jak bardzo przodek żeński
        (np. matka) jest spokrewniony do danej osoby
        """

        if inlaw == '':
            t_inlaw = ""
        else:
            t_inlaw = "przybrana "

        if level >= 0 and level < len(_mother_level):
            # Jeśli znasz bezpośrednią nazwę relacji, to ją zastosuj

            if level == 1:
                # matka
                return t_inlaw + _mother_level[level]

            elif (level >= 2) & (level <= 4):
                # babcia, prababcia, praprababcia
                return t_inlaw + _mother_level[level] \
                    + self.get_sword_distaff(level, reltocommon, ' ')

            else:
                return t_inlaw + _mother_level[level]

        elif level >= len(_mother_level) and (level - 1) < len(_level_name):
            # jeśli istnieje liczebnik dla danej liczby
            return t_inlaw + \
                    "przodek żeński %s pokolenia" % (_level_name[level - 1])

        else:
            # dla pozostałych przypadków wypisz relację liczbowo
            return t_inlaw +"przodek żeński w %d pokoleniu" % level



    def get_parent_unknown(self, level, inlaw=''):
        """
        Podaje tekst zawierający informację, jak bardzo przodek
        o nieokreślonej płci jest spokrewniony dodanej osoby
        """

        if inlaw == '':
            t_inlaw = ""
        else:
            t_inlaw = "przybrany "

        if level == 1:
            return t_inlaw + "rodzic"

        elif level > 1 and (level - 1) < len(_level_name):
            if (level >= 2) & (level <= 4):
                # babcia, prababcia, praprababcia
                # (albo dziadek, pradziadek, prapradziadek)
                tmp = t_inlaw +\
                    "przodek %s pokolenia" % (_level_name[level - 1])
                # TODO: try to recognize a gender...
                return tmp
                # +  self.get_sword_distaff(level, reltocommon, ' ')
            else:
                return t_inlaw + \
                   "przodek %s pokolenia" % (_level_name[level - 1])
        else:
            return t_inlaw +"przodek w %d pokoleniu" % level


    def get_single_relationship_string(self, Ga, Gb, gender_a, gender_b,
                                       reltocommon_a, reltocommon_b,
                                       only_birth=True,
                                       in_law_a=False, in_law_b=False):
        """
        Provide a string that describes the relationsip between a person, and
        another person. E.g. "grandparent" or "child".
        """

        if only_birth:
            step = ''
        else:
            step = self.STEP

        if in_law_a or in_law_b :
            inlaw = self.INLAW
        else:
            inlaw = ''


        # b is the same person as a
        if Ga == Gb == 0:
            rel_str = 'ta sama osoba'

        elif Ga == 0:
            # b is son/descendant of a

            if gender_b == Person.MALE:
                if inlaw and Gb == 1 and not step:
                    rel_str = "zięć"
                else:
                    rel_str = self.get_son(Gb, inlaw)

            elif gender_b == Person.FEMALE:
                if inlaw and Gb == 1 and not step:
                    rel_str = "synowa"
                else:
                    rel_str = self.get_daughter(Gb, inlaw)

            else:
                rel_str = self.get_child_unknown(Gb, inlaw)


        elif Gb == 0:
            # b is parent/grand parent of a

            if gender_b == Person.MALE:
                if inlaw and Gb == 1 and not step:
                    # TODO: znaleźć odpowiedniki w zależności czy to syn/córka
                    rel_str = "teść"
                else:
                    rel_str = self.get_father(Ga, reltocommon_a, inlaw)

            elif gender_b == Person.FEMALE:
                if inlaw and Gb == 1 and not step:
                    # TODO: znaleźć odpowiedniki w zależności czy to syn/córka
                    rel_str = "teściowa"
                else:
                    rel_str = self.get_mother(Ga, reltocommon_a, inlaw)

            else:
                rel_str = self.get_parent_unknown(Ga, inlaw)


        elif Ga == Gb == 1:
            # rodzeństwo
            if gender_b == Person.MALE:
                if inlaw and not step:
                    rel_str = "brat przyrodni"
                else:
                    rel_str = "brat rodzony"

            elif gender_b == Person.FEMALE:
                if inlaw and not step:
                    rel_str = "siostra przyrodnia"
                else:
                    rel_str = "siostra rodzony"
            else:
                rel_str = "brat/siostra"

        elif Gb == 1 and Ga > 1:

            # Przyjmij, że nie rozróżniamy osób prawnie i nieprawnie przybranych...

            if Ga == 2:
                # rodzeństwo rodziców

                # brat ojca, czyli stryj
                if (gender_b == Person.MALE) \
                    & (reltocommon_a[0] == self.REL_FATHER):
                    rel_str = "stryj"

                # siostra ojca, czyli ciotka ???
                elif (gender_b == Person.FEMALE) \
                    & (reltocommon_a[0] == self.REL_FATHER):
                    rel_str = "ciotka (tzw. stryjna)"

                # brat matki, czyli wuj/wujek
                elif (gender_b == Person.MALE) \
                    & (reltocommon_a[0] == self.REL_MOTHER):
                    rel_str = "wuj (wujek)"

                # siostra matki, czyli ciotka
                elif (gender_b == Person.FEMALE) \
                    & (reltocommon_a[0] == self.REL_MOTHER):
                    rel_str = "ciotka"

                else:
                    rel_str = "brat lub siostra rodzica"

            elif Ga == 3:
                # rodzeństwo dziadków rodziców osoby sprawdzanej

                # rodzeństwo dziadka po mieczu (ojca ojca)
                if (reltocommon_a[0] == self.REL_FATHER) \
                    & (reltocommon_a[1] == self.REL_FATHER):

                    if  (gender_b == Person.MALE):
                        rel_str = "dziadek stryjeczny (tzw przestryj, stary stryj)"
                    elif (gender_b == Person.FEMALE):
                        rel_str = "babcia stryjeczna"
                    else:
                        rel_str = "rodzeństwo przodka w 2 pokoleniu"

                # rodzeństwo babki po mieczu (matki ojca)
                elif (reltocommon_a[0] == self.REL_FATHER) \
                    & (reltocommon_a[1] == self.REL_MOTHER):

                    # TODO: Należy sprawdzić, czy w staropolszczyźnie nie ma
                    #            dokładniejszych określeń dla tego typu relacji
                    # TODO: EN: Try to check, whether in old polish language
                    #      are more specific word for this kind of relation
                    if  (gender_b == Person.MALE):
                        rel_str = "dziadek stryjeczny (tzw przestryj, stary stryj)"
                    elif (gender_b == Person.FEMALE):
                        rel_str = "babcia stryjeczna"
                    else:
                        rel_str = "rodzeństwo przodka w 2 pokoleniu"

                # rodzeństwo dziadka po kądzieli (ojca matki)
                elif (reltocommon_a[0] == self.REL_MOTHER) \
                    & (reltocommon_a[1] == self.REL_FATHER):

                    # TODO: Należy sprawdzić, czy w staropolszczyźnie nie ma
                    #            dokładniejszych określeń dla tego typu relacji
                    # TODO: EN: Try to check, whether in old polish language
                    #      are more specific word for this kind of relation
                    if  (gender_b == Person.MALE):
                        rel_str = "dziadek cioteczny (starop. prapociot)"
                    elif (gender_b == Person.FEMALE):
                        rel_str = "babcia cioteczna (starop. praciota)"
                    else:
                        rel_str = "rodzeństwo przodka w 2 pokoleniu"


                # rodzeństwo babki po kądzieli (matki matki)
                elif (reltocommon_a[0] == self.REL_MOTHER) \
                    & (reltocommon_a[1] == self.REL_MOTHER):

                    # TODO: Należy sprawdzić, czy w staropolszczyźnie nie ma
                    #           dokładniejszych określeń dla tego typu relacji
                    # TODO: EN: Try to check, whether in old polish language
                    #      are more specific word for this kind of relation
                    if  (gender_b == Person.MALE):
                        rel_str = "dziadek cioteczny (starop. prapociot)"
                    elif (gender_b == Person.FEMALE):
                        rel_str = "babcia cioteczna  (starop. praciota)"
                    else:
                        rel_str = "rodzeństwo przodka w 2 pokoleniu"

                else:
                    if  (gender_b == Person.MALE):
                        rel_str = "rodzeństwo dziadka"
                    elif (gender_b == Person.FEMALE):
                        rel_str = "rodzeństwo babci"
                    else:
                        rel_str = "rodzeństwo przodka w 2 pokoleniu"

            elif Ga > 3:
                # pradziadkowie...  (grandparents)

                if (gender_b == Person.MALE) \
                    & (reltocommon_a[0] == self.REL_FATHER):

                    if Ga >= 0 and Ga < len(_brother_level_of_male):
                        rel_str = _brother_level_of_male[Ga]
                    else:
                        rel_str = "rodzeństwo przodka męskiego %d pokolenia" % Ga

                elif (gender_b == Person.FEMALE) \
                    & (reltocommon_a[0] == self.REL_FATHER):
                    if Ga >= 0 and Ga < len(_sister_level_of_male):
                        rel_str = _sister_level_of_male[Ga]
                    else:
                        rel_str = "rodzeństwo przodka żeńskiego %d pokolenia" % Ga

                elif (gender_b == Person.MALE) \
                    & (reltocommon_a[0] == self.REL_MOTHER):

                    if Ga >= 0 and Ga < len(_brother_level_of_female):
                        rel_str = _brother_level_of_male[Ga]
                    else:
                        rel_str = "rodzeństwo przodka męskiego %d pokolenia" % Ga

                elif (gender_b == Person.FEMALE) \
                    & (reltocommon_a[0] == self.REL_MOTHER):
                    if Ga >= 0 and Ga < len(_sister_level_of_female):
                        rel_str = _sister_level_of_male[Ga]
                    else:
                        rel_str = "rodzeństwo przodka żeńskiego %d pokolenia" % Ga

                else:
                    rel_str = "rodzeństwo przodka %d pokolenia" % Ga
            else:
                # A program should never goes there, but...
                rel_str = "Relacja nie określona"

        elif Ga  ==1 and Gb > 1:

            # syn brata
            if (gender_b == Person.MALE) \
                & (reltocommon_b[0] == self.REL_FATHER):
                    if Gb < len(_nephew_level_of_brothers_son):
                        rel_str = _nephew_level_of_brothers_son[Gb]
                    else:
                        rel_str = "męski potomek w %d pokoleniu brata" % Gb

            # córka brata
            elif (gender_b == Person.FEMALE) \
                & (reltocommon_b[0] == self.REL_FATHER):
                    if Gb < len(_nephew_level_of_brothers_daughter):
                        rel_str = _nephew_level_of_brothers_daughter[Gb]
                    else:
                        rel_str = "żeński potomek w %d pokoleniu brata" % Gb

            # syn siostry
            if (gender_b == Person.MALE) \
                & (reltocommon_b[0] == self.REL_MOTHER):
                    if Gb < len(_nephew_level_of_sisters_son):
                        rel_str = _nephew_level_of_sisters_son[Gb]
                    else:
                        rel_str = "męski potomek w %d pokoleniu brata" % Gb

            # córka siostry
            elif (gender_b == Person.FEMALE) \
                & (reltocommon_b[0] == self.REL_MOTHER):
                    if Gb < len(_nephew_level_of_sisters_daughter):
                        rel_str = _nephew_level_of_sisters_daughter[Gb]
                    else:
                        rel_str = "żeński potomek w %d pokoleniu brata" % Gb

            # potomek brata
            elif (reltocommon_b[0] == self.REL_FATHER):
                rel_str = "potomek w %d pokoleniu brata" % Gb

            # potomek brata
            elif (reltocommon_b[0] == self.REL_MOTHER):
                rel_str = "potomek w %d pokoleniu brata" % Gb

            else :
                rel_str = "potomek w %d pokoleniu rodzeństwa" % Gb

        elif Ga > 1 and Gb > 1:
            if (gender_b == Person.MALE):
                if Ga == 2 and Gb == 2:
                    rel_str = "kuzyn"
                else:
                    rel_str = "daleki kuzyn (%d. stopień pokrewieństwa)" % (Ga+Gb)

            elif (gender_b == Person.FEMALE):
                if Ga == 2 and Gb == 2:
                    rel_str = "kuzynka"
                else:
                    rel_str = "daleka kuzynka (%d. stopień pokrewieństwa)" % (Ga+Gb)

            else:
                if Ga == 2 and Gb == 2:
                    rel_str = "kuzyn(ka)"
                else:
                    rel_str = "daleki członek rodziny (%d. stopień pokrewieństwa)" % (Ga+Gb)

        else:
            # A program should never goes there, but...
            rel_str = "nieokreślony stopień pokrewieństwa"

        return rel_str

    def get_plural_relationship_string(self, Ga, Gb,
                                       reltocommon_a='', reltocommon_b='',
                                       only_birth=True,
                                       in_law_a=False, in_law_b=False):
        """
        Generate a text with information, how far away is a group of persons
        from a main person
        """

        if Ga == Gb == 0:
            return 'ta sama osoba'
        if 0 == Ga:
            if 1 == Gb:
                return 'Dzieci'
            if 2 == Gb:
                return 'Wnuki'
            if 3 == Gb:
                return 'Prawnuki'
            if 4 == Gb:
                return 'Praprawnuki'
            return 'Praprapra(n)wnuki'
        if 0 == Gb:
            if 1 == Ga:
                return 'Rodzice'
            if 2 == Ga:
                return 'Dziadkowie'
            if 3 == Ga:
                return 'Pradziadkowie'
            if 4 == Ga:
                return 'Praprapradziadkowie'
            return 'Praprapra(n)dziadkowie'
        if 1 == Ga == Gb:
            return 'Rodzeństwo'
        if 1 == Gb and Ga > 1:
            return 'Wujowie/stryjowie i ciocie'
        if 1 < Gb and 1 == Ga:
            return 'bratankowie(ice)/siostrzeńcy(nice)'
        if 1 < Ga and 1 < Gb:
            return 'dalsza rodzina'
        return 'relacja nieznana'

    def get_sibling_relationship_string(self, sib_type, gender_a, gender_b,
                                        in_law_a=False, in_law_b=False):

        if in_law_a or in_law_b :
            inlaw = self.INLAW
        else:
            inlaw = ''

        if sib_type == self.NORM_SIB:
            if not inlaw:
                if gender_b == Person.MALE:
                    rel_str = 'brat (rodzony)'
                elif gender_b == Person.FEMALE:
                    rel_str = 'siostra (rodzona)'
                else:
                    rel_str = 'brat lub siostra (rodzeni)'
            else:
                if gender_b == Person.MALE:
                    # TODO: znaleźć odpowiednik
                    rel_str = "brat (pasierb)"
                elif gender_b == Person.FEMALE:
                    # TODO: znaleźć odpowiednik
                    rel_str = "siostra (pasierbica)"
                else:
                    # TODO: znaleźć odpowiednik
                    rel_str = "brat lub siostra (pasierb/pasierbica)"
        elif sib_type == self.UNKNOWN_SIB:
            if not inlaw:
                if gender_b == Person.MALE:
                    rel_str = 'brat'
                elif gender_b == Person.FEMALE:
                    rel_str = 'siostra'
                else:
                    rel_str = 'brat lub siostra'
            else:
                if gender_b == Person.MALE:
                    # TODO: znaleźć odpowiednik
                    rel_str = "brat (brat/szwagier)"
                elif gender_b == Person.FEMALE:
                    # TODO: znaleźć odpowiednik
                    rel_str = "siostra (bratowa/szwagierka)"
                else:
                    # TODO: znaleźć odpowiednik
                    rel_str = "brat lub siostra (szwagier/szagierka)"
        elif sib_type == self.HALF_SIB_FATHER:
                if gender_b == Person.MALE:
                    rel_str = "brat przyrodni"
                elif gender_b == Person.FEMALE:
                    rel_str = "siostra przyrodnia"
                else:
                    rel_str = "brat/siostra przyrodni"
        elif sib_type == self.HALF_SIB_MOTHER:
                if gender_b == Person.MALE:
                    rel_str = "brat przyrodni"
                elif gender_b == Person.FEMALE:
                    rel_str = "siostra przyrodnia"
                else:
                    rel_str = "brat/siostra przyrodni"
        elif sib_type == self.STEP_SIB:
                if gender_b == Person.MALE:
                    rel_str = "brat przyrodni"
                elif gender_b == Person.FEMALE:
                    rel_str = "siostra przyrodnia"
                else:
                    rel_str = "brat lub siostra przyrodnia"
        else:
            rel_str = "nieokreślona relacja rodzeństwa"
        return rel_str


if __name__ == "__main__":
    # Test function. Call it as follows from the command line (so as to find
    #        imported modules):
    #    export PYTHONPATH=/path/to/gramps/src
    #    python src/plugins/rel/rel_pl.py

    """TRANSLATORS, copy this if statement at the bottom of your
        rel_xx.py module, and test your work with:
        python src/plugins/rel/rel_xx.py
    """
    from gramps.gen.relationship import test
    RC = RelationshipCalculator()
    test(RC, True)



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
# Copyright (C) 2014       Mathieu MD <mathieu.md@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
#

"""
Japanese-specific classes for parsing and displaying dates.
"""

#-------------------------------------------------------------------------
#
# Python modules
#
#-------------------------------------------------------------------------

import re

#-------------------------------------------------------------------------
#
# Gramps modules
#
#-------------------------------------------------------------------------

from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler

#-------------------------------------------------------------------------
#
# Japanese parser
#
#-------------------------------------------------------------------------
class DateParserJA(DateParser):
    """
    Convert a text string into a Date object. If the date cannot be
    converted, the text string is assigned.
    """

    # modifiers before the date
    modifier_to_int = {
        '以前' : Date.MOD_BEFORE,
        '以降' : Date.MOD_AFTER,
        '頃'   : Date.MOD_ABOUT,
        'ごろ' : Date.MOD_ABOUT,
        }

    month_to_int = DateParser.month_to_int

    quality_to_int = {
        'およそ'    : Date.QUAL_ESTIMATED,
        'ごろ'      : Date.QUAL_ESTIMATED,
        '位'      : Date.QUAL_ESTIMATED,
        '計算上'  : Date.QUAL_CALCULATED,
        }

    bce = ["紀元前", "BC"] + DateParser.bce

    def init_strings(self):
        """
        This method compiles regular expression strings for matching dates.
        """
        DateParser.init_strings(self)

        DateParser.calendar_to_int.update({
            'グレゴリオ暦' : Date.CAL_GREGORIAN,
            'g'            : Date.CAL_GREGORIAN,
            'ユリウス暦' : Date.CAL_JULIAN,
            'j'          : Date.CAL_JULIAN,
            'ユダヤ暦' : Date.CAL_HEBREW,
            'h'        : Date.CAL_HEBREW,
            'ヒジュラ暦' : Date.CAL_ISLAMIC,
            'i'          : Date.CAL_ISLAMIC,
            'フランス革命暦' : Date.CAL_FRENCH,
            '共和暦'         : Date.CAL_FRENCH,
            'f'              : Date.CAL_FRENCH,
            'イラン暦' : Date.CAL_PERSIAN,
            'p'        : Date.CAL_PERSIAN,
            'スウェーデン暦' : Date.CAL_SWEDISH,
            's'              : Date.CAL_SWEDISH,
        })

        DateParser.month_to_int.update({
            "一月"      : 1,
            "ichigatsu" : 1,
            "睦月"      : 1,
            "mutsuki"   : 1,
            "二月"       : 2,
            "nigatsu"    : 2,
            "如月"       : 2,
            "kisaragi"   : 2,
            "衣更着"     : 2,
            "kinusaragi" : 2,
            "三月"     : 3,
            "sangatsu" : 3,
            "弥生"     : 3,
            "yayoi"    : 3,
            "四月"     : 4,
            "shigatsu" : 4,
            "卯月"     : 4,
            "uzuki"    : 4,
            "五月"       : 5,
            "gogatsu"    : 5,
            "皐月"       : 5,
            "satsuki"    : 5,
            "早苗月"     : 5,
            "sanaetsuki" : 5,
            "六月"      : 6,
            "rokugatsu" : 6,
            "水無月"    : 6,
            "minazuki"  : 6,
            "七月"        : 7,
            "shichigatsu" : 7,
            "文月"        : 7,
            "fumizuki"    : 7,
            "八月"       : 8,
            "hachigatsu" : 8,
            "葉月"       : 8,
            "hazuki"     : 8,
            "九月"      : 9,
            "kugatsu"   : 9,
            "長月"      : 9,
            "nagatsuki" : 9,
            "十月"        : 10,
            "jugatsu"     : 10,
            "jūgatsu"     : 10,
            "juugatsu"    : 10,
            "神無月"      : 10,
            "kannazuki"   : 10,
            "kaminazuki"  : 10,
            "神有月"      : 10,
            "神在月"      : 10,
            "kamiarizuki" : 10,
            "十一月"       : 11,
            "juichigatsu"  : 11,
            "jūichigatsu"  : 11,
            "juuichigatsu" : 11,
            "霜月"         : 11,
            "shimotsuki"   : 11,
            "十二月"     : 12,
            "junigatsu"  : 12,
            "jūnigatsu"  : 12,
            "juunigatsu" : 12,
            "師走"       : 12,
            "shiwasu"    : 12,
        })

        _span_1 = ['から', '~', '〜']
        _span_2 = ['まで', '']
        _range_1 = ['から', 'と', '~', '〜']
        _range_2 = ['までの間', 'の間']
        self._span = re.compile(r"(?P<start>.+)(%s)(?P<stop>\d+)(%s)" %
                                ('|'.join(_span_1), '|'.join(_span_2)),
                                re.IGNORECASE)
        self._range = re.compile(r"(?P<start>.+)(%s)(?P<stop>.+)(%s)" %
                                 ('|'.join(_range_1), '|'.join(_range_2)),
                                 re.IGNORECASE)
        self._numeric = re.compile(r"((\d+)年\s*)?((\d+)月\s*)?(\d+)?日?\s*$")

#-------------------------------------------------------------------------
#
# Japanese display
#
#-------------------------------------------------------------------------
class DateDisplayJA(DateDisplay):
    """
    Japanese language date display class.
    """

    def formats_changed(self):
        """ Allow overriding so a subclass can modify """

        # Specify what is actually the "System Default".
        example = self.dhformat
        example = example.replace('%d', "31")
        example = example.replace('%m', "12")
        example = example.replace('%Y', "1999")

        # This definition must agree with its "_display_gregorian" method
        self. formats = ("YYYY-MM-DD (ISO)", # 0
                "システムデフォールト (" + example + ")", # 1
                "1999年12月31日",   # 2
                "1999年十二月31日", # 3
                )

    def _display_gregorian(self, date_val, **kwargs):
        """
        display gregorian calendar date in different format
        """
        # this must agree with its locale-specific "formats" definition
        year = self._slash_year(date_val[2], date_val[3])
        if self.format == 0:
            # ISO
            return self.display_iso(date_val)

        elif self.format == 1:
            # System Default
            if date_val[2] < 0 or date_val[3]:
                return self.display_iso(date_val)
            else:
                if date_val[0] == date_val[1] == 0:
                    value = str(date_val[2])
                else:
                    value = self.dhformat.replace('%m', str(date_val[1]))
                    if date_val[0] == 0: # ignore the zero day and its delimiter
                        i_day = value.find('%d')
                        value = value.replace(value[i_day:i_day+3], '')
                    value = value.replace('%d', str(date_val[0]))
                    value = value.replace('%Y', str(date_val[2]))

        elif self.format == 2:
            # 1999年12月31日
            if date_val[0] == 0:
                if date_val[1] == 0:
                    value = "%s年" % year
                else:
                    value = "%s年%s" % (year,
                                        self.short_months[date_val[1]])
            else:
                value = "%s年%s%s日" % (year,
                                        self.short_months[date_val[1]],
                                        date_val[0])

        elif self.format == 3:
            # 1999年十二月31日
            if date_val[0] == 0:
                if date_val[1] == 0:
                    value = "%s年" % year
                else:
                    value = "%s年%s" % (year,
                                        self.long_months[date_val[1]])
            else:
                value = "%s年%s%s日" % (year,
                                        self.long_months[date_val[1]],
                                        date_val[0])

        else:
            return self.display_iso(date_val)

        if date_val[2] < 0:
            return self._bce_str % value
        else:
            return value

    display = DateDisplay.display_formatted

#-------------------------------------------------------------------------
#
# Register classes
#
#-------------------------------------------------------------------------

register_datehandler(
    ('ja_JP', 'ja', 'japanese', 'Japanese', ('%Y年%m月%d日',)),
    DateParserJA, DateDisplayJA)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007-2012  Brian G. Matherly
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2011       Matt Keenan (matt.keenan@gmail.com)
# Copyright (C) 2013-2014  Paul Franklin
# Copyright (C) 2010,2015  Craig J. Anderson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Reports/Text Reports/Descendant Report.
"""

#------------------------------------------------------------------------
#
# standard python modules
#
#------------------------------------------------------------------------

#------------------------------------------------------------------------
#
# Gramps modules
#
#------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
_ = glocale.translation.gettext
from gramps.gen.plug.docgen import (IndexMark, FontStyle, ParagraphStyle,
                                    FONT_SANS_SERIF, INDEX_TYPE_TOC,
                                    PARA_ALIGN_CENTER)
from gramps.gen.plug.menu import (NumberOption, PersonOption, BooleanOption,
                                  EnumeratedListOption)
from gramps.gen.errors import ReportError
from gramps.gen.plug.report import Report
from gramps.gen.plug.report import utils
from gramps.gen.plug.report import MenuReportOptions
from gramps.gen.plug.report import stdoptions
from gramps.gen.utils.db import (get_birth_or_fallback, get_death_or_fallback,
                                 get_marriage_or_fallback,
                                 get_divorce_or_fallback)
from gramps.gen.proxy import CacheProxyDb
from gramps.gen.display.place import displayer as _pd
from gramps.gen.display.name import displayer as _nd

#------------------------------------------------------------------------
#
# PrintDAboville
#
#------------------------------------------------------------------------
class PrintDAboville():
    """
    d'Aboville numbering system

    (according to en.wikipedia.org/Genealogical_numbering_systems
    his name is spelled "d'Aboville" and not "D'Aboville" but I will
    leave this class name alone, mainly fixing the translated string,
    so that it is both accurate and also agrees with the DDR string)
    """

    def __init__(self):
        self.num = [0]

    def number(self, level):
        """ Make the current number based upon the current level """
        # Set up the array based on the current level
        while len(self.num) > level:  # We can go from a level 8 to level 2
            self.num.pop()
        if len(self.num) < level:
            self.num.append(0)

        # Increment the current level - initalized with 0
        self.num[-1] += 1

        # Display
        return ".".join(map(str, self.num))


#------------------------------------------------------------------------
#
# PrintHenry
#
#------------------------------------------------------------------------
class PrintHenry():
    """ Henry numbering system """

    def __init__(self, modified=False):
        self.henry = "123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        self.modified = modified
        self.num = [0]

    def number(self, level):
        """ Make the current number based upon the current level """
        # Set up the array based on the current level
        while len(self.num) > level:  # We can go from a level 8 to level 2
            self.num.pop()
        if len(self.num) < level:
            self.num.append(0)

        # Incriment the current level - initalized with 0
        self.num[-1] += 1

        def strd(inti):
            """ no change needed """
            return "(" + str(inti) + ")"

        # Display
        if self.modified is False:
            return "".join(map(
                lambda x: self.henry[x-1] if x <= len(self.henry) else strd(x)
                , self.num))
        else:
            return "".join(map(
                lambda x: str(x) if x < 10 else strd(x)
                , self.num))


#------------------------------------------------------------------------
#
# PrintSimple
#
#------------------------------------------------------------------------
class PrintSimple:
    """ Simple numbering system """

    def __init__(self, showdups):
        self.showdups = showdups
        self.num = {0:1}

    def number(self, level):
        """ the number of the person """
        if self.showdups:
            # Just show original simple numbering
            to_return = "%d." % level
        else:
            to_return = str(level)
            if level > 1:
                to_return += "-" + str(self.num[level-1])
            to_return += "."

            self.num[level] = 1
            self.num[level-1] = self.num[level-1] + 1

        return to_return


#------------------------------------------------------------------------
#
# PrintVlliers
#   de_Villiers_Pama numbering system
#
#------------------------------------------------------------------------
class PrintVilliers:
    """ de_Villiers_Pama numbering system """

    def __init__(self):
        self.pama = 'abcdefghijklmnopqrstuvwxyz'
        self.num = {0:1}

    def number(self, level):
        """ the number of the person """
        to_return = self.pama[level-1]
        if level > 1:
            to_return += str(self.num[level-1])
        to_return += "."

        self.num[level] = 1
        self.num[level-1] = self.num[level-1] + 1

        return to_return


#------------------------------------------------------------------------
#
# class PrintMeurgey
#   Meurgey_de_Tupigny numbering system
#
#------------------------------------------------------------------------
class PrintMeurgey:
    """ Meurgey_de_Tupigny numbering system """

    def __init__(self):
        self.childnum = [""]

    def number(self, level):
        """ the number of the person """
        if level == 1:
            dash = ""
        else:
            dash = "-"
            if len(self.childnum) < level:
                self.childnum.append(1)

        to_return = (utils.roman(level) + dash +
                     str(self.childnum[level-1]) + ".")

        if level > 1:
            self.childnum[level-1] += 1

        return to_return


#------------------------------------------------------------------------
#
# Printinfo
#
#------------------------------------------------------------------------
class Printinfo:
    """
    A base class used to help make the individual numbering system classes.
    This class must first be initialized with set_class_vars
    """
    def __init__(self, doc, database, numbering, showmarriage, showdivorce,
                 name_display, rlocale, want_ids, pformat):
        #classes
        self._name_display = name_display
        self.doc = doc
        self.database = database
        self.numbering = numbering
        #variables
        self.showmarriage = showmarriage
        self.showdivorce = showdivorce
        self.want_ids = want_ids
        self._ = rlocale.translation.sgettext # needed for English
        self._get_date = rlocale.get_date
        self.pformat = pformat

    def __date_place(self, event):
        """ return the date and/or place an event happened """
        if event:
            date = self._get_date(event.get_date_object())
            place_handle = event.get_place_handle()
            if place_handle:
                place = _pd.display_event(self.database, event, self.pformat)
                return("%(event_abbrev)s %(date)s - %(place)s" % {
                    'event_abbrev': event.type.get_abbreviation(self._),
                    'date' : date,
                    'place' : place,
                    })
            else:
                return("%(event_abbrev)s %(date)s" % {
                    'event_abbrev': event.type.get_abbreviation(self._),
                    'date' : date
                    })
        return ""

    def dump_string(self, person, family=None):
        """ generate a descriptive string for a person """
        string = self.__date_place(
            get_birth_or_fallback(self.database, person))

        tmp = self.__date_place(
            get_death_or_fallback(self.database, person))
        if string and tmp:
            string += self._(", ") # Arabic OK
        string += tmp

        if string:
            string = " (" + string + ")"

        if family and self.showmarriage:
            tmp = self.__date_place(
                get_marriage_or_fallback(self.database, family))
            if tmp:
                string += self._(", ") + tmp # Arabic OK

        if family and self.showdivorce:
            tmp = self.__date_place(
                get_divorce_or_fallback(self.database, family))
            if tmp:
                string += self._(", ") + tmp # Arabic OK

        if family and self.want_ids:
            string += ' (%s)' % family.get_gramps_id()

        self.doc.write_text(string)

    def print_person(self, level, person):
        """ print the person """
        display_num = self.numbering.number(level)
        self.doc.start_paragraph("DR-Level%d" % min(level, 32), display_num)
        mark = utils.get_person_mark(self.database, person)
        self.doc.write_text(self._name_display.display(person), mark)
        if self.want_ids:
            self.doc.write_text(' (%s)' % person.get_gramps_id())
        self.dump_string(person)
        self.doc.end_paragraph()
        return display_num

    def print_spouse(self, level, spouse_handle, family_handle):
        """ print the spouse """
        #Currently print_spouses is the same for all numbering systems.
        if spouse_handle:
            spouse = self.database.get_person_from_handle(spouse_handle)
            mark = utils.get_person_mark(self.database, spouse)
            self.doc.start_paragraph("DR-Spouse%d" % min(level, 32))
            name = self._name_display.display(spouse)
            self.doc.write_text(
                self._("sp. %(spouse)s") % {'spouse':name}, mark)
            if self.want_ids:
                self.doc.write_text(' (%s)' % spouse.get_gramps_id())
            self.dump_string(spouse, family_handle)
            self.doc.end_paragraph()
        else:
            self.doc.start_paragraph("DR-Spouse%d" % min(level, 32))
            self.doc.write_text(
                self._("sp. %(spouse)s") % {'spouse':self._('Unknown')})
            self.doc.end_paragraph()

    def print_reference(self, level, person, display_num):
        """ print the reference """
        #Person and their family have already been printed so
        #print reference here
        if person:
            mark = utils.get_person_mark(self.database, person)
            self.doc.start_paragraph("DR-Spouse%d" % min(level, 32))
            name = self._name_display.display(person)
            self.doc.write_text(self._("sp. see %(reference)s: %(spouse)s"
                                      ) % {'reference' : display_num,
                                           'spouse'    : name},
                                mark)
            self.doc.end_paragraph()


#------------------------------------------------------------------------
#
# RecurseDown
#
#------------------------------------------------------------------------
class RecurseDown:
    """
    A simple object to recurse from a person down through their descendants

    The arguments are:

    max_generations: The max number of generations
    database:  The database object
    obj_print: A Printinfo derived class that prints person
               information on the report
    """
    def __init__(self, max_generations, database,
                 obj_print, showdups, rlocale):
        self.max_generations = max_generations
        self.database = database
        self.obj_print = obj_print
        self.showdups = showdups
        self.person_printed = {}
        self._ = rlocale.translation.sgettext # needed for English

    def recurse(self, level, person, curdepth):
        """ recurse """

        person_handle = person.get_handle()
        display_num = self.obj_print.print_person(level, person)

        if curdepth is None:
            ref_str = display_num
        else:
            ref_str = curdepth + " " + display_num

        if person_handle not in self.person_printed:
            self.person_printed[person_handle] = ref_str

        for family_handle in person.get_family_handle_list():
            family = self.database.get_family_from_handle(family_handle)

            spouse_handle = utils.find_spouse(person, family)

            if not self.showdups and spouse_handle in self.person_printed:
                # Just print a reference
                spouse = self.database.get_person_from_handle(spouse_handle)
                self.obj_print.print_reference(
                    level, spouse, self.person_printed[spouse_handle])
            else:
                self.obj_print.print_spouse(level, spouse_handle, family)

                if spouse_handle:
                    spouse_num = self._("%s sp.") % ref_str
                    self.person_printed[spouse_handle] = spouse_num

                if level >= self.max_generations:
                    continue

                childlist = family.get_child_ref_list()[:]
                for child_ref in childlist:
                    child = self.database.get_person_from_handle(child_ref.ref)
                    self.recurse(level+1, child, ref_str)


#------------------------------------------------------------------------
#
# DescendantReport
#
#------------------------------------------------------------------------
class DescendantReport(Report):
    """ Descendant report """

    def __init__(self, database, options, user):
        """
        Create the DescendantReport object that produces the report.

        The arguments are:

        database        - the Gramps database instance
        options         - instance of the Options class for this report
        user            - a gen.user.User() instance

        This report needs the following parameters (class variables)
        that come in the options class.

        gen           - Maximum number of generations to include.
        name_format   - Preferred format to display names
        dups          - Whether to include duplicate descendant trees
        incl_private  - Whether to include private data
        living_people - How to handle living people
        years_past_death - Consider as living this many years after death
        inc_id        - Whether to include Gramps IDs
        """

        Report.__init__(self, database, options, user)

        menu = options.menu

        self.set_locale(menu.get_option_by_name('trans').get_value())

        stdoptions.run_date_format_option(self, menu)

        stdoptions.run_private_data_option(self, menu)
        stdoptions.run_living_people_option(self, menu, self._locale)
        self.database = CacheProxyDb(self.database)

        self.max_generations = menu.get_option_by_name('gen').get_value()
        self.want_ids = menu.get_option_by_name('inc_id').get_value()

        pid = menu.get_option_by_name('pid').get_value()
        self.center_person = self.database.get_person_from_gramps_id(pid)
        if self.center_person is None:
            raise ReportError(_("Person %s is not in the Database") % pid)

        #Initialize the Printinfo class
        self._showdups = menu.get_option_by_name('dups').get_value()
        numbering = menu.get_option_by_name('numbering').get_value()
        if numbering == "Simple":
            obj = PrintSimple(self._showdups)
        elif numbering == "Henry":
            obj = PrintHenry()
        elif numbering == "Modified Henry":
            obj = PrintHenry(modified=True)
        elif numbering == "d'Aboville":
            obj = PrintDAboville()
        elif numbering == "de Villiers/Pama":
            obj = PrintVilliers()
        elif numbering == "Meurgey de Tupigny":
            obj = PrintMeurgey()
        else:
            raise AttributeError("no such numbering: '%s'" % numbering)

        marrs = menu.get_option_by_name('marrs').get_value()
        divs = menu.get_option_by_name('divs').get_value()

        stdoptions.run_name_format_option(self, menu)

        pformat = menu.get_option_by_name("place_format").get_value()

        self.obj_print = Printinfo(self.doc, self.database, obj, marrs, divs,
                                   self._name_display, self._locale,
                                   self.want_ids, pformat)

    def write_report(self):
        self.doc.start_paragraph("DR-Title")
        name = self._name_display.display(self.center_person)
        # feature request 2356: avoid genitive form
        title = self._("Descendants of %s") % name
        mark = IndexMark(title, INDEX_TYPE_TOC, 1)
        self.doc.write_text(title, mark)
        self.doc.end_paragraph()

        recurse = RecurseDown(self.max_generations, self.database,
                              self.obj_print, self._showdups, self._locale)
        recurse.recurse(1, self.center_person, None)

#------------------------------------------------------------------------
#
# DescendantOptions
#
#------------------------------------------------------------------------
class DescendantOptions(MenuReportOptions):

    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, dbase):
        self.__db = dbase
        self.__pid = None
        MenuReportOptions.__init__(self, name, dbase)

    def get_subject(self):
        """ Return a string that describes the subject of the report. """
        gid = self.__pid.get_value()
        person = self.__db.get_person_from_gramps_id(gid)
        return _nd.display(person)

    def add_menu_options(self, menu):
        category_name = _("Report Options")

        self.__pid = PersonOption(_("Center Person"))
        self.__pid.set_help(_("The center person for the report"))
        menu.add_option(category_name, "pid", self.__pid)

        numbering = EnumeratedListOption(_("Numbering system"), "Simple")
        numbering.set_items([
            ("Simple", _("Simple numbering")),
            ("d'Aboville", _("d'Aboville numbering")),
            ("Henry", _("Henry numbering")),
            ("Modified Henry", _("Modified Henry numbering")),
            ("de Villiers/Pama", _("de Villiers/Pama numbering")),
            ("Meurgey de Tupigny", _("Meurgey de Tupigny numbering"))])
        numbering.set_help(_("The numbering system to be used"))
        menu.add_option(category_name, "numbering", numbering)

        gen = NumberOption(_("Generations"), 10, 1, 15)
        gen.set_help(_("The number of generations to include in the report"))
        menu.add_option(category_name, "gen", gen)

        stdoptions.add_gramps_id_option(menu, category_name)

        marrs = BooleanOption(_('Show marriage info'), False)
        marrs.set_help(
            _("Whether to show marriage information in the report."))
        menu.add_option(category_name, "marrs", marrs)

        divs = BooleanOption(_('Show divorce info'), False)
        divs.set_help(_("Whether to show divorce information in the report."))
        menu.add_option(category_name, "divs", divs)

        dups = BooleanOption(_('Show duplicate trees'), True)
        dups.set_help(
            _("Whether to show duplicate Family Trees in the report."))
        menu.add_option(category_name, "dups", dups)

        category_name = _("Report Options (2)")

        stdoptions.add_name_format_option(menu, category_name)

        stdoptions.add_place_format_option(menu, category_name)

        stdoptions.add_private_data_option(menu, category_name)

        stdoptions.add_living_people_option(menu, category_name)

        locale_opt = stdoptions.add_localization_option(menu, category_name)

        stdoptions.add_date_format_option(menu, category_name, locale_opt)

    def make_default_style(self, default_style):
        """Make the default output style for the Descendant Report."""
        fstyle = FontStyle()
        fstyle.set_size(12)
        fstyle.set_type_face(FONT_SANS_SERIF)
        fstyle.set_bold(1)
        pstyle = ParagraphStyle()
        pstyle.set_header_level(1)
        pstyle.set_bottom_border(1)
        pstyle.set_top_margin(utils.pt2cm(3))
        pstyle.set_bottom_margin(utils.pt2cm(3))
        pstyle.set_font(fstyle)
        pstyle.set_alignment(PARA_ALIGN_CENTER)
        pstyle.set_description(_("The style used for the title."))
        default_style.add_paragraph_style("DR-Title", pstyle)

        fstyle = FontStyle()
        fstyle.set_size(10)
        for i in range(1, 33):
            pstyle = ParagraphStyle()
            pstyle.set_font(fstyle)
            pstyle.set_top_margin(utils.pt2cm(fstyle.get_size()*0.125))
            pstyle.set_bottom_margin(
                utils.pt2cm(fstyle.get_size()*0.125))
            pstyle.set_first_indent(-0.5)
            pstyle.set_left_margin(min(10.0, float(i-0.5)))
            pstyle.set_description(
                _("The style used for the level %d display.") % i)
            default_style.add_paragraph_style("DR-Level%d" % min(i, 32),
                                              pstyle)

            pstyle = ParagraphStyle()
            pstyle.set_font(fstyle)
            pstyle.set_top_margin(utils.pt2cm(fstyle.get_size()*0.125))
            pstyle.set_bottom_margin(
                utils.pt2cm(fstyle.get_size()*0.125))
            pstyle.set_left_margin(min(10.0, float(i-0.5)))
            pstyle.set_description(
                _("The style used for the spouse level %d display.") % i)
            default_style.add_paragraph_style("DR-Spouse%d" % min(i, 32),
                                              pstyle)
